import os
import re
import glob
import shutil
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
from pathlib import Path
from natsort import natsorted
from functools import partial
from astropy.table import Table
from concurrent.futures import ProcessPoolExecutor

from ..utils import data_utils
from ..utils import logging_utils
from ..paper.superNNova_plots import datasets_plots


def process_fn(inputs):
    fn, fil = inputs
    return fn(fil)


def powers_of_two(x):
    powers = []
    i = 1
    while i <= x:
        if i & x:
            powers.append(i)
        i <<= 1
    return powers


def build_traintestval_splits(settings):
    """Build dataset split in the following way

    - Downsample each class so that it has the same cardinality as the lowest cardinality class
    - Randomly assign lightcurves to a 80/10/10 train test val split (except Out-of-distribution data 1/1/98)

    OOD:
        Will use the complete sample for testing, does not require settings.

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    logging_utils.print_green("Computing splits")

    # Read and process files faster with ProcessPoolExecutor
    max_workers = multiprocessing.cpu_count() - 2
    photo_columns = ["SNID"] + [
        f"target_{nb_classes}classes"
        # for nb_classes in list(set([2, len(settings.sntypes.keys())]))
        for nb_classes in list(
            set([2, len(set([k for k in dict(settings.sntypes).values()]))])
        )
    ]

    # Load headers
    # either in HEAD.FITS or csv format
    list_files = natsorted(Path(settings.raw_dir).glob("**/*HEAD*"))
    list_fmt = [re.search(r"(FITS|csv)", fil.name).group() for fil in list_files]
    list_files = [str(fil) for fil in list_files]

    print("List files", list_files)
    # use parallelization to speed up processing
    if not settings.debug:
        process_fn_FITS = partial(
            data_utils.process_header_FITS,
            settings=settings,
            columns=photo_columns + [settings.sntype_var],
        )
        process_fn_csv = partial(
            data_utils.process_header_csv,
            settings=settings,
            columns=photo_columns + [settings.sntype_var],
        )

        list_fn = []
        for fmt in list_fmt:
            if fmt == "csv":
                list_fn.append(process_fn_csv)
            elif fmt == "FITS":
                list_fn.append(process_fn_FITS)

        list_pairs = list(zip(list_fn, list_files))
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            list_df = executor.map(process_fn, list_pairs)

    else:
        logging_utils.print_yellow("Beware debugging mode (loop over files)")
        list_df = []
        for fil, fmat in zip(list_files, list_fmt):
            if fmat == "FITS":
                list_df.append(
                    data_utils.process_header_FITS(
                        fil, settings, columns=photo_columns + [settings.sntype_var]
                    )
                )
            else:
                list_df.append(
                    data_utils.process_header_csv(
                        fil, settings, columns=photo_columns + [settings.sntype_var]
                    )
                )

    # Load df_photo
    df_photo = pd.concat(list_df)
    df_photo["SNID"] = df_photo["SNID"].astype(str).str.strip()
    # load FITOPT file on which we will base our splits
    df_salt = data_utils.load_fitfile(settings)
    if len(df_salt) < 1:
        # if no fits file we include all lcs
        logging_utils.print_yellow("All lcs used for salt and photometry samples")
        df_salt = pd.DataFrame()
        df_salt["SNID"] = df_photo["SNID"].str.strip()
    df_salt["is_salt"] = 1
    # correct format SNID
    df_salt["SNID"] = df_salt["SNID"].astype(str).str.strip()
    # Check all SNID in df_salt are also in df_photo
    try:
        assert np.all(df_salt.SNID.isin(df_photo.SNID))
    except Exception:
        logging_utils.print_red(
            " BEWARE! This is not the fits file for this photometry "
        )
        print(logging_utils.str_to_redstr("   do point at the correct --fits_dir "))
        print(
            logging_utils.str_to_redstr(
                "   (cheat: or an empty folder to override use of salt2fits) "
            )
        )
        import sys

        sys.exit(1)
    # Merge left on df_photo
    df = df_photo.merge(df_salt[["SNID", "is_salt"]], on=["SNID"], how="left")
    # Drop duplicates in case of multiple entries for the same SNID
    df = df.drop_duplicates(subset="SNID")
    # Some curves are in photo and not in salt, these curves have is_salt = NaN
    # We replace the NaN with 0
    df["is_salt"] = df["is_salt"].fillna(0).astype(int)

    # Save dataset stats
    list_stat = []

    # Save a dataframe to record train/test/val split for
    # binary, ternary and all-classes classification
    for dataset in ["saltfit", "photometry"]:
        for nb_classes in list(
            set([2, len(set([k for k in dict(settings.sntypes).values()]))])
        ):
            logging_utils.print_green(
                f"Computing {dataset} splits for {nb_classes}-way classification"
            )
            # Randomly sample SNIDs such that all class have the same number of occurences
            if dataset == "saltfit":
                g = df[df.is_salt == 1].groupby(
                    f"target_{nb_classes}classes", group_keys=False
                )
            else:
                g = df.groupby(f"target_{nb_classes}classes", group_keys=False)
            dic_targets = (
                g[settings.sntype_var].apply(lambda x: list(np.unique(x))).to_dict()
            )
            print(f"target {settings.sntype_var}")
            settings.data_types_training = [
                (
                    f"{k} {settings.sntypes[v[0]]} {[int(dt) for dt in dic_targets[k]]}"
                    if v[0] in settings.sntypes.keys()
                    else f"{k} other {[int(dt) for dt in dic_targets[k]]}"
                )
                for k, v in dic_targets.items()
            ]
            print(settings.data_types_training)

            if settings.testing_ids:
                if Path(settings.testing_ids).suffix == ".csv":
                    df_ids_test = pd.read_csv(settings.testing_ids)
                    try:
                        ids_test = df_ids_test["SNID"].astype(str).values
                    except Exception:
                        logging_utils.print_red(
                            f"Provide a {settings.testing_ids} with SNID column"
                        )
                        raise ValueError
                elif Path(settings.testing_ids).suffix == ".npy":
                    ids_test = np.load(settings.testing_ids)
                    ids_test = [f"{k}" for k in ids_test]
                else:
                    logging_utils.print_red("Provide a csv or numpy testing_ids file")
                    raise ValueError

                g_wo_test = df[~df.SNID.isin(ids_test)].groupby(
                    f"target_{nb_classes}classes", group_keys=False
                )
                g_test = df[df.SNID.isin(ids_test)].groupby(
                    f"target_{nb_classes}classes", group_keys=False
                )
            # Line below: we have grouped df by target, we find out which of those
            # group has the smallest size with g.size().min(), then we sample randomly
            # from this group and reset the index. We then sample with frac=1 to shuffle
            # the whole dataset. Otherwise, the classes are sorted and the train/test/val
            # splits are incorrect.
            if settings.data_testing:
                # when just classifying data balancing is not necessary
                g = g.apply(lambda x: x).reset_index(drop=True).sample(frac=1)
            elif settings.testing_ids:
                g_wo_test = (
                    g_wo_test.apply(lambda x: x).reset_index(drop=True).sample(frac=1)
                )
                g_test = g_test.apply(lambda x: x).reset_index(drop=True).sample(frac=1)
            else:
                g = (
                    g.apply(lambda x: x.sample(g.size().min()))
                    .reset_index(drop=True)
                    .sample(frac=1)
                )
            if settings.testing_ids:
                sampled_SNIDs_wo_test = g_wo_test["SNID"].values
                n_samples = len(sampled_SNIDs_wo_test)
                SNID_train = sampled_SNIDs_wo_test[: int(0.9 * n_samples)]
                SNID_val = sampled_SNIDs_wo_test[int(0.9 * n_samples) : int(n_samples)]
                sampled_SNIDs_test = g_test["SNID"].values
                SNID_test = sampled_SNIDs_test[:]
            else:
                sampled_SNIDs = g["SNID"].values
                n_samples = len(sampled_SNIDs)
                # Now create train/test/validation indices
                if settings.data_training:
                    SNID_train = sampled_SNIDs[: int(0.99 * n_samples)]
                    SNID_val = sampled_SNIDs[
                        int(0.99 * n_samples) : int(0.995 * n_samples)
                    ]
                    SNID_test = sampled_SNIDs[int(0.995 * n_samples) :]
                elif settings.data_testing:
                    SNID_test = sampled_SNIDs[:]
                    # the train and val sets wont be used in this case
                    SNID_train = [sampled_SNIDs[0]]
                    SNID_val = [sampled_SNIDs[0]]
                else:
                    SNID_train = sampled_SNIDs[: int(0.8 * n_samples)]
                    SNID_val = sampled_SNIDs[
                        int(0.8 * n_samples) : int(0.9 * n_samples)
                    ]
                    SNID_test = sampled_SNIDs[int(0.9 * n_samples) :]

            # Find the indices of our train test val splits
            idxs_train = np.where(df.SNID.isin(SNID_train))[0]
            idxs_val = np.where(df.SNID.isin(SNID_val))[0]
            idxs_test = np.where(df.SNID.isin(SNID_test))[0]
            # Create a new column that will state to which data split
            # a given SNID will be assigned
            # train: 0, val: 1, test:2 others: -1
            arr_dataset = -np.ones(len(df)).astype(int)
            arr_dataset[idxs_train] = 0
            arr_dataset[idxs_val] = 1
            arr_dataset[idxs_test] = 2

            df[f"dataset_{dataset}_{nb_classes}classes"] = arr_dataset
            # Display classes balancing in the dataset, for each split
            logging_utils.print_bright("Dataset composition")
            for split_name, idxs in zip(
                ["Training", "Validation", "Test"], [idxs_train, idxs_val, idxs_test]
            ):
                # We count the number of occurence in each class and each split with
                # pandas.value_counts
                d_occurences = (
                    df[f"target_{nb_classes}classes"]
                    .iloc[idxs]
                    .value_counts()
                    .sort_values()
                    .to_dict()
                )
                d_occurences_SNTYPE = (
                    df[settings.sntype_var]
                    .iloc[idxs]
                    .value_counts()
                    .sort_values()
                    .to_dict()
                )
                total_samples = sum(d_occurences.values())
                total_samples_str = logging_utils.str_to_yellowstr(total_samples)

                str_ = f"# samples {total_samples_str} "
                for c_, n_samples in d_occurences.items():
                    class_str = logging_utils.str_to_yellowstr(c_)
                    class_fraction = f"{100 *(n_samples/total_samples):.2g}%"
                    class_fraction_str = logging_utils.str_to_yellowstr(class_fraction)
                    str_ += f"Class {class_str}: {class_fraction_str} samples "
                list_stat.append(
                    [
                        dataset,
                        nb_classes,
                        split_name,
                        total_samples,
                        d_occurences,
                        d_occurences_SNTYPE,
                    ]
                )

                logging_utils.print_green(f"{split_name} set", str_)

    # Save to pickle
    df.to_pickle(f"{settings.processed_dir}/SNID.pickle", protocol=4)

    logging_utils.print_green("Done")


def process_single_FITS(file_path, settings):
    """
    Carry out preprocessing on FITS file and save results to pickle.
    Pickle is preferred to csv as it is faster to read and write.

    - Join column from header files
    - Select columns that will be useful laer on
    - Compute SNID to tag each light curve
    - Compute delta times between measures
    - Filter preprocessing
    - Removal of delimiter rows

    Args:
        file_path (str): path to ``.FITS`` file
        settings (ExperimentSettings): controls experiment hyperparameters

    """
    # Load the PHOT file
    df = data_utils.load_pandas_from_fit(file_path)

    if len(df) < 1:
        logging_utils.print_red("Do not provide empty photometry file", file_path)
    # Last line may be a line with MJD = -777.
    # Remove it so that it does not interfere with arr_ID below
    if df.MJD.values[-1] == -777.0:
        df = df.drop(df.index[-1])

    # Keep only columns of interest
    keep_col = ["MJD", "FLUXCAL", "FLUXCALERR", "FLT"]
    # BAND and FLT are exchangeable
    if "FLT" not in df.keys() and "BAND" in df.keys():
        df = df.rename(columns={"BAND": "FLT"})
    df = (
        df[keep_col + [settings.phot_reject]].copy()
        if settings.phot_reject
        else df[keep_col].copy()
    )
    # filters have a trailing white space which we remove and convert to string
    df.FLT = df.FLT.apply(
        lambda x: (
            x.decode("utf-8").rstrip() if isinstance(x, bytes) else str(x).rstrip()
        )
    )
    # Load the companion HEAD file
    header = Table.read(file_path.replace("PHOT", "HEAD"), format="fits")
    df_header = header.to_pandas()
    try:
        df_header["SNID"] = df_header["SNID"].str.decode("utf-8")
    except Exception:
        df_header["SNID"] = df_header["SNID"].astype(str)
    # Keep only columns of interest
    # Hack for using the final redshift not the galaxy
    if settings.redshift_label != "none":
        logging_utils.print_yellow("Changed redshift to", settings.redshift_label)
        df_header["HOSTGAL_SPECZ"] = df_header[settings.redshift_label]
        df_header["HOSTGAL_SPECZ_ERR"] = df_header[f"{settings.redshift_label}_ERR"]
        df_header["SIM_REDSHIFT_CMB"] = df_header[settings.redshift_label]

    keep_col_header = [
        "SNID",
        "PEAKMJD",
        "HOSTGAL_PHOTOZ",
        "HOSTGAL_PHOTOZ_ERR",
        "HOSTGAL_SPECZ",
        "HOSTGAL_SPECZ_ERR",
        "SIM_REDSHIFT_CMB",
        "SIM_PEAKMAG_z",
        "SIM_PEAKMAG_g",
        "SIM_PEAKMAG_r",
        "SIM_PEAKMAG_i",
        settings.sntype_var,
    ]
    if settings.photo_window_var not in keep_col_header:
        keep_col_header += [settings.photo_window_var]
    if settings.additional_train_var:
        keep_col_header += list(settings.additional_train_var)
    # check if keys are in header
    keep_col_header = [k for k in keep_col_header if k in df_header.keys()]
    df_header = df_header[keep_col_header].copy()
    df_header["SNID"] = df_header["SNID"].astype(str).str.strip()

    #############################################
    # Photometry window init
    #############################################
    if settings.photo_window_files:
        if Path(settings.photo_window_files[0]).exists():
            # load fits file
            df_peak = pd.read_csv(
                settings.photo_window_files[0],
                comment="#",
                delimiter=" ",
                skipinitialspace=True,
            )
            if "SNID" not in df_peak.keys():
                df_peak["SNID"] = df_peak["CID"].astype(str)
            else:
                df_peak["SNID"] = df_peak["SNID"].astype(str)

            try:
                df_peak = df_peak[["SNID", settings.photo_window_var]]
            except Exception:
                logging_utils.print_red("Provide a correct photo_window variable")
                raise Exception
            # merge with header
            df_header_tmp = pd.merge(df_header, df_peak, on="SNID")
            if len(df_header) == len(df_header_tmp):
                df_header = df_header_tmp
            else:
                raise Exception
            if len(df_header) < 1:
                logging_utils.print_red(
                    "Provide a matching photo_window_file (not a common SNID found) "
                )
                raise Exception
        else:
            if settings.photo_window_files[0] == "HEAD":
                # if using a variable from header file
                if settings.photo_window_var in df_header.keys():
                    pass
                else:
                    logging_utils.print_red(
                        "Provide a valid peak key in header or a photo_window_file"
                    )
                    logging_utils.print_red(
                        f"Currently {settings.photo_window_var} {settings.photo_window_files}"
                    )
            else:
                logging_utils.print_red(
                    "Provide a valid peak key in header or a photo_window_file"
                )
                logging_utils.print_red(
                    f"Currently {settings.photo_window_var} {settings.photo_window_files}"
                )
    #############################################
    # Compute SNID for df and join with df_header
    #############################################
    arr_ID = np.chararray(len(df), itemsize=15)
    # New light curves are identified by MJD == -777.0
    arr_idx = np.where(df["MJD"].values == -777.0)[0]
    arr_idx = np.hstack((np.array([0]), arr_idx, np.array([len(df)])))
    # Fill in arr_ID
    for counter in range(1, len(arr_idx)):
        start, end = arr_idx[counter - 1], arr_idx[counter]
        # index starts at zero
        arr_ID[start:end] = df_header.SNID.iloc[counter - 1]
    df["SNID"] = arr_ID.astype(str)
    df["SNID"] = df["SNID"].str.strip()
    df = df.set_index("SNID")
    df_header["SNID"] = df_header["SNID"].str.strip()
    df_header = df_header.set_index("SNID")
    # join df and header
    df = df.join(df_header).reset_index()

    #############################################
    # Photometry window & quality (flag) selection
    #############################################
    # window
    if settings.photo_window_files:
        df["window_time_cut"] = True
        mask = df["MJD"] != -777.00
        df["window_delta_time"] = df["MJD"] - df[settings.photo_window_var]
        df.loc[mask, "window_time_cut"] = df["window_delta_time"].apply(
            lambda x: (
                True
                if (x > 0 and x < settings.photo_window_max)
                else (True if (x <= 0 and x > settings.photo_window_min) else False)
            )
        )
        df = df[df["window_time_cut"]]
    # quality
    if settings.phot_reject:
        # only valid for powers of two combinations
        tmp = len(df.SNID.unique())
        tmp2 = len(df)
        df["phot_reject"] = df[settings.phot_reject].apply(
            lambda x: (
                False
                if len(
                    set(settings.phot_reject_list).intersection(set(powers_of_two(x)))
                )
                > 0
                else True
            )
        )
        df = df[df["phot_reject"]]

        if settings.debug:
            logging_utils.print_blue("Phot reject", file_path)
            logging_utils.print_blue(f"SNID {tmp} to {len(df.SNID.unique())}")
            logging_utils.print_blue(f"Phot {tmp2} to {len(df)}")

    #############################################
    # Miscellaneous data processing
    #############################################
    df = df[keep_col + keep_col_header].copy()

    # keep only filters we are going to use for classification
    df = df[df["FLT"].isin(settings.list_filters)]
    if len(df) < 1:
        logging_utils.print_red(
            "No light curve left after filtering by filters, check your settings"
        )
        raise ValueError

    # Drop the delimiter lines
    df = df[df.MJD != -777.000]
    # Reset the index (it is no longer continuous after dropping lines)
    df = df.reset_index(drop=True)
    # Add delta time
    df = data_utils.compute_delta_time(df)
    # Remove rows post large delta time in the same light curve(delta_time > 150)
    # df = data_utils.remove_data_post_large_delta_time(df)
    #############################################
    # Add class and dataset information
    #############################################
    df_SNID = pd.read_pickle(f"{settings.processed_dir}/SNID.pickle")
    # Check all SNID in df are in df_SNID
    assert np.all(df.SNID.isin(df_SNID.SNID))
    # Merge left on df: len(df) will not change and will now include
    # relevant columns from df_SNID
    merge_columns = ["SNID"]
    # for c_ in list(set([2, len(settings.sntypes.keys())])):
    distinct_classes = len(set([k for k in dict(settings.sntypes).values()]))
    for c_ in list(set([2, distinct_classes])):
        merge_columns += [f"target_{c_}classes"]
        for dataset in ["photometry", "saltfit"]:
            merge_columns += [f"dataset_{dataset}_{c_}classes"]
    df = df.merge(df_SNID[merge_columns], on=["SNID"], how="left")

    # Save for future use
    basename = os.path.basename(file_path)
    folder_name = Path(file_path.split(f"{settings.raw_dir}/")[-1]).parent
    if folder_name != Path("."):
        prefix = str(folder_name).replace("/", "_")
        basename = f"{prefix}_{basename}"

    df.to_pickle(
        f"{settings.preprocessed_dir}/{basename.replace('.FITS', '.pickle').replace('.gz','')}",
        protocol=4,
    )

    # getting SNIDs for SNe with Host_spec
    host_spe = df[df["HOSTGAL_SPECZ"] > 0]["SNID"].unique().tolist()

    print(len(df), "rows in preprocessed file", file_path)
    return host_spe


def process_single_csv(file_path, settings):
    """
    Carry out preprocessing on csv file and save results to pickle.
    Pickle is preferred to csv as it is faster to read and write.

    - Compute delta times between measures
    - Filter preprocessing

    Args:
        file_path (str): path to ``.csv`` file
        settings (ExperimentSettings): controls experiment hyperparameters

    """
    # Load the PHOT file
    df = pd.read_csv(file_path)
    if len(df) < 1:
        logging_utils.print_red("Do not provide empty photometry file", file_path)
        raise ValueError

    # Keep only columns of interest
    keep_col = ["SNID", "MJD", "FLUXCAL", "FLUXCALERR", "FLT"]
    df = df[keep_col].copy()
    df["SNID"] = df["SNID"].astype(str)
    df = df.set_index("SNID")

    # Load the companion HEAD file
    df_header = pd.read_csv(file_path.replace("PHOT", "HEAD"))

    if settings.redshift_label != "none":
        logging_utils.print_yellow("Changed redshift to", settings.redshift_label)
        df_header["HOSTGAL_SPECZ"] = df_header[settings.redshift_label]
        df_header["HOSTGAL_SPECZ_ERR"] = df_header[f"{settings.redshift_label}_ERR"]
        df_header["SIM_REDSHIFT_CMB"] = df_header[settings.redshift_label]

    # Keep only columns of interest
    keep_col_header = [
        "SNID",
        "PEAKMJD",
        "HOSTGAL_PHOTOZ",
        "HOSTGAL_PHOTOZ_ERR",
        "HOSTGAL_SPECZ",
        "HOSTGAL_SPECZ_ERR",
        "SIM_REDSHIFT_CMB",
        "SIM_PEAKMAG_z",
        "SIM_PEAKMAG_g",
        "SIM_PEAKMAG_r",
        "SIM_PEAKMAG_i",
        settings.sntype_var,
    ]
    if settings.photo_window_var not in keep_col_header:
        keep_col_header += [settings.photo_window_var]
    if settings.additional_train_var:
        keep_col_header += list(settings.additional_train_var)
        print(f"Adding additional variables to dataset {settings.additional_train_var}")
    # check if keys are in header
    keep_col_header = [k for k in keep_col_header if k in df_header.keys()]
    df_header = df_header[keep_col_header].copy()
    df_header["SNID"] = df_header["SNID"].astype(str)
    df_header["SNID"] = df_header["SNID"].str.strip()
    df_header = df_header.set_index("SNID")
    df = df.join(df_header).reset_index()

    if settings.photo_window_files:
        df["window_time_cut"] = True
        mask = df["MJD"] != -777.00
        df["window_delta_time"] = df["MJD"] - df[settings.photo_window_var]
        df.loc[mask, "window_time_cut"] = df["window_delta_time"].apply(
            lambda x: (
                True
                if (x > 0 and x < settings.photo_window_max)
                else (True if (x <= 0 and x > settings.photo_window_min) else False)
            )
        )
        df = df[df["window_time_cut"] is True]
    # quality
    if settings.phot_reject:

        tmp = len(df.SNID.unique())
        tmp2 = len(df)
        df["phot_reject"] = df[settings.phot_reject].apply(
            lambda x: (
                False
                if len(
                    set(settings.phot_reject_list).intersection(set(powers_of_two(x)))
                )
                > 0
                else True
            )
        )
        df = df[df["phot_reject"] is True]

        if settings.debug:
            logging_utils.print_blue("Phot reject", file_path)
            logging_utils.print_blue(f"SNID {tmp} to {len(df.SNID.unique())}")
            logging_utils.print_blue(f"Phot {tmp2} to {len(df)}")

    #############################################
    # Miscellaneous data processing
    #############################################
    df = df[list(set(keep_col + keep_col_header))].copy()
    # filters have a trailing white space which we remove
    df.FLT = df.FLT.apply(lambda x: x.rstrip()).values.astype(str)
    # keep only filters we are going to use for classification
    df = df[df["FLT"].isin(settings.list_filters)]
    # Drop the delimiter lines
    df = df[df.MJD != -777.000]
    # Reset the index (it is no longer continuous after dropping lines)
    df = df.reset_index(drop=True)
    # Add delta time
    df = data_utils.compute_delta_time(df)
    # Remove rows post large delta time in the same light curve(delta_time > 150)
    # df = data_utils.remove_data_post_large_delta_time(df)

    #############################################
    # Add class and dataset information
    #############################################
    df_SNID = pd.read_pickle(f"{settings.processed_dir}/SNID.pickle")
    # Check all SNID in df are in df_SNID
    assert np.all(df.SNID.isin(df_SNID.SNID))
    # Merge left on df: len(df) will not change and will now include
    # relevant columns from df_SNID
    merge_columns = ["SNID"]
    # for c_ in list(set([2, len(settings.sntypes.keys())])):
    distinct_classes = len(set([k for k in dict(settings.sntypes).values()]))
    for c_ in list(set([2, distinct_classes])):
        merge_columns += [f"target_{c_}classes"]
        for dataset in ["photometry", "saltfit"]:
            merge_columns += [f"dataset_{dataset}_{c_}classes"]
    df = df.merge(df_SNID[merge_columns], on=["SNID"], how="left")

    # Save for future use
    basename = os.path.basename(file_path)
    df.to_pickle(
        f"{settings.preprocessed_dir}/{basename.replace('.csv', '.pickle').replace('.gz','')}",
        protocol=4,
    )

    # getting SNIDs for SNe with Host_spec
    host_spe = (
        df[df["HOSTGAL_SPECZ"] > 0]["SNID"].unique().tolist()
        if "HOSTGAL_SPECZ" in df.keys()
        else []
    )

    return host_spe


def preprocess_data(settings):
    """Preprocess the FITS data

    - Use multiprocessing/threading to speed up data processing
    - Preprocess every FIT file in the raw data dir
    - Also save a DataFrame of Host Spe for publication plots

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters

    """

    # Get the list of FITS files
    # Load headers
    # either in HEAD.FITS or csv format
    list_files = natsorted(Path(settings.raw_dir).glob("**/*PHOT*"))
    list_fmt = [re.search(r"(FITS|csv)", fil.name).group() for fil in list_files]
    list_files = [str(fil) for fil in list_files]

    if not settings.debug:
        process_fn_FITS = partial(process_single_FITS, settings=settings)
        process_fn_csv = partial(process_single_csv, settings=settings)

        list_fn = []
        for fmt in list_fmt:
            if fmt == "csv":
                list_fn.append(process_fn_csv)
            elif fmt == "FITS":
                list_fn.append(process_fn_FITS)

    logging_utils.print_green("List to preprocess ", list_files)
    max_workers = multiprocessing.cpu_count() - 2

    host_spe_tmp = []
    # use parallelization to speed up processing
    # Split list files in chunks of size 10 or less
    # to get a progress bar and alleviate memory constraints
    num_elem = len(list_files)
    num_chunks = num_elem // 10 + 1
    list_chunks = np.array_split(np.arange(num_elem), num_chunks)

    # # Loop over chunks of files
    if not settings.debug:
        for chunk_idx in tqdm(list_chunks, desc="Preprocess", ncols=100):
            # Process each file in the chunk in parallel
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                start, end = chunk_idx[0], chunk_idx[-1] + 1
                # Need to cast to list because executor returns an iterator
                # host_spe_tmp += list(executor.map(parallel_fn, list_files[start:end]))
                list_pairs = list(zip(list_fn[start:end], list_files[start:end]))
                host_spe_tmp += list(executor.map(process_fn, list_pairs))

    else:
        logging_utils.print_yellow("Beware debugging mode (loop over files)")
        # for debugging only (parallelization needs to be commented)
        for i in range(len(list_files)):
            out = (
                process_single_FITS(list_files[i], settings)
                if "FITS" in list_files[i]
                else process_single_csv(list_files[i], settings)
            )
            host_spe_tmp.append(out)
    # Save host spe for plotting and performance tests
    host_spe = [item for sublist in host_spe_tmp for item in sublist]
    pd.DataFrame(host_spe, columns=["SNID"]).to_pickle(
        f"{settings.processed_dir}/hostspe_SNID.pickle", protocol=4
    )
    logging_utils.print_green("Finished preprocessing")


def pivot_dataframe_single(filename, settings):

    df = pd.read_pickle(filename)
    df = pivot_dataframe_single_from_df(df, settings)

    # Save to pickle
    dump_filename = filename.split(".pickle")[0] + "_pivot.pickle"
    df.to_pickle(dump_filename, protocol=4)


def pivot_dataframe_single_from_df(df, settings):
    """
    Carry out pivot: we will group time-wise close observations on the same row
    and each row in the dataframe will show a value for each of the flux and flux
    error column

    - All observations withing 8 hours of each other are assigned the same MJD
    - Results are cached with pickle

    Args:
        filename (str): path to a ``.pickle`` file containing pre-processed data
        settings (ExperimentSettings): controls experiment hyperparameters

    """
    list_filters = settings.list_filters

    assert len(list_filters) > 0

    arr_MJD = df.MJD.values
    arr_delta_time = df.delta_time.values
    # Loop over times to create grouped MJD:
    # if filters are acquired within less than 0.33 MJD (~8 hours) of each other
    # they get assigned the same time
    min_dt = 0.33
    time_last_change = 0
    arr_grouped_MJD = np.zeros_like(arr_MJD)
    for i in range(len(df)):
        time = arr_MJD[i]
        dt = arr_delta_time[i]
        # 2 possibilities to update the time
        # dt == 0 (it"s a new light curve)
        # time - time_last_change > min_dt
        if dt == 0 or (time - time_last_change) > min_dt:
            arr_grouped_MJD[i] = time
            time_last_change = time
        else:
            arr_grouped_MJD[i] = arr_grouped_MJD[i - 1]
    # Add grouped delta time to dataframe
    df["grouped_MJD"] = np.array(arr_grouped_MJD)

    # Some filters (i, r, g, z) may appear multiple times with the same grouped MJD within same light curve
    # When this happens, we select the one with lowest FLUXCALERR
    df = df.sort_values("FLUXCALERR").groupby(["SNID", "grouped_MJD", "FLT"]).first()
    # We then reset the index
    df = df.reset_index()
    # Compute PEAKMJDNORM = PEAKMJD in days since the start of the light curve
    df["PEAKMJDNORM"] = df["PEAKMJD"] - df["MJD"]
    # The correct PEAKMJDNORM is the first one hence the use of first after groupby
    df_PEAKMJDNORM = df[["SNID", "PEAKMJDNORM"]].groupby("SNID").first().reset_index()
    # Remove PEAKMJDNORM
    df = df.drop(columns="PEAKMJDNORM")
    # Add PEAKMJDNORM back to df with a merge on SNID
    df = df.merge(df_PEAKMJDNORM, how="left", on="SNID")
    # drop columns that won"t be used onwards
    df = df.drop(columns=["MJD", "delta_time"])
    class_columns = []
    # for c_ in list(set([2, len(settings.sntypes.keys())])):
    distinct_classes = len(set([k for k in dict(settings.sntypes).values()]))
    for c_ in list(set([2, distinct_classes])):
        class_columns += [f"target_{c_}classes"]
        for dataset in ["photometry", "saltfit"]:
            class_columns += [f"dataset_{dataset}_{c_}classes"]
    group_features_list = (
        [
            "SNID",
            "grouped_MJD",
            "PEAKMJD",
            "PEAKMJDNORM",
            "SIM_REDSHIFT_CMB",
            settings.sntype_var,
            "SIM_PEAKMAG_z",
            "SIM_PEAKMAG_g",
            "SIM_PEAKMAG_r",
            "SIM_PEAKMAG_i",
        ]
        + [k for k in df.keys() if "HOST" in k]
        + class_columns
    )
    if settings.photo_window_var not in group_features_list:
        group_features_list += [settings.photo_window_var]
    if settings.additional_train_var:
        group_features_list += list(settings.additional_train_var)
    # check if keys are in header
    group_features_list = [k for k in group_features_list if k in df.keys()]
    # Pivot so that for a given MJD, we have info on all available fluxes / error
    df = pd.pivot_table(df, index=group_features_list, columns=["FLT"])

    # Flatten columns
    df.columns = ["_".join(col).strip() for col in df.columns.values]
    # Reset index to get grouped_MJD and target as columns
    cols_to_reset_list = [c for c in df.index.names if c != "SNID"]
    df = df.reset_index(cols_to_reset_list)
    # Rename grouped_MJD to MJD
    df = df.rename(columns={"grouped_MJD": "MJD"})

    # New column to indicate which channel (r,g,z,i) is present
    # The column will read ``rg`` if r,g are present; ``rgz`` if r,g and z are present, etc.
    # fix missing filters
    missing_filters = [k for k in list_filters if f"FLUXCAL_{k}" not in df.columns]
    for f in missing_filters:
        df[f"FLUXCAL_{f}"] = np.nan
        df[f"FLUXCALERR_{f}"] = np.nan
    for flt in list_filters:
        df[flt] = np.where(df["FLUXCAL_%s" % flt].isnull(), "", flt)
    df["FLT"] = df[list_filters[0]]
    for flt in list_filters[1:]:
        df["FLT"] += df[flt]
    # Drop some irrelevant columns
    df = df.drop(columns=list_filters)
    # Finally replace NaN with 0
    df = df.fillna(0)
    # Add delta_time back. We removed all delta time columns above as they get
    # filled with NaN during pivot. It is clearer to recompute delta time once the pivot is complete
    df = data_utils.compute_delta_time(df)

    # Cast columns to float32, int32 to save space
    for c in df.columns:
        if df[c].dtype == np.float64:
            df[c] = df[c].astype(np.float32)
        elif "classes" in c and df[c].dtype == np.int64:
            df[c] = df[c].astype(np.int8)
    # Add some extra columns from the FITOPT file
    df_salt = data_utils.load_fitfile(settings, verbose=False)
    if len(df_salt) > 1:
        df_salt = df_salt.set_index("SNID")
    else:
        # if no fits file we populate with dummies
        # logging_utils.print_yellow(f"Creating dummy mB,c,x1")
        df_salt = pd.DataFrame()
        df_salt["SNID"] = np.array(df.index.unique())
        df_salt["mB"] = np.zeros(len(df.index.unique()))
        df_salt["c"] = np.zeros(len(df.index.unique()))
        df_salt["x1"] = np.zeros(len(df.index.unique()))
        df_salt = df_salt.set_index("SNID")
    df = df.join(df_salt[["mB", "c", "x1"]], how="left")

    df = df.drop(columns="MJD")

    return df


def pivot_dataframe_batch(list_files, settings):
    """
    - Use multiprocessing/threading to speed up data processing
    - Pivot every file in list_files and cache the result with pickle

    Args:
        list_files (list): list of ``.pickle`` files containing pre-processed data
        settings (ExperimentSettings): controls experiment hyperparameters

    """
    # Split list files in chunks of size 10 or less
    # to get a progress bar and alleviate memory constraints
    num_elem = len(list_files)
    num_chunks = num_elem // 10 + 1
    list_chunks = np.array_split(np.arange(num_elem), num_chunks)

    # Parameters of multiprocessing below
    if not settings.debug:
        max_workers = multiprocessing.cpu_count() - 2
        # use parallelization to speed up processing
        # Loop over chunks of files
        for chunk_idx in tqdm(list_chunks, desc="Pivoting dataframes", ncols=100):
            parallel_fn = partial(pivot_dataframe_single, settings=settings)
            # Process each file in the chunk in parallel
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                start, end = chunk_idx[0], chunk_idx[-1] + 1
                executor.map(parallel_fn, list_files[start:end])
    else:
        logging_utils.print_yellow("Beware debugging mode (loop over pivot)")
        # for debugging only, process one file only
        for fil in list_files:
            pivot_dataframe_single(fil, settings)

    logging_utils.print_green("Finished pivot")


def detect_contaminant_types(settings):
    """Pre-scan raw data files to detect types not in settings.sntypes.

    Any type found in the data but missing from settings.sntypes is
    automatically added as 'contaminant'. This must run before any
    column-name computation so that target column names are consistent
    throughout the pipeline.

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """
    list_files = natsorted(Path(settings.raw_dir).glob("**/*HEAD*"))
    all_types = set()
    for fil in list_files:
        fmt = re.search(r"(FITS|csv)", fil.name).group()
        if fmt == "csv":
            df_head = pd.read_csv(fil, usecols=[settings.sntype_var])
        else:
            df_head = Table.read(str(fil), format="fits").to_pandas()
        all_types.update(df_head[settings.sntype_var].astype(str).unique())

    missing_types = [t for t in sorted(all_types) if t not in settings.sntypes]
    if missing_types:
        logging_utils.print_yellow(
            "Missing sntypes",
            f"{missing_types} assigned to 'contaminant' class",
        )
        for mtyp in missing_types:
            settings.sntypes[mtyp] = "contaminant"

    # Warn about sntypes keys not found in the data but do NOT remove them.
    # Phantom keys are harmless: groupby only creates groups for values present
    # in the data, so downsampling and splits work correctly. Keeping them
    # preserves the class structure (target_Nclasses column name and indices),
    # which is essential for compatibility when using a model trained on a
    # dataset that contained all types.
    phantom_keys = [k for k in list(settings.sntypes.keys()) if k not in all_types]
    if phantom_keys:
        logging_utils.print_yellow(
            "Unused sntypes",
            f"Keys {phantom_keys} not found in data (kept for class structure consistency)",
        )


@logging_utils.timer("Data processing")
def make_dataset(settings):
    """Main function for data processing

    - Create the train test val splits
    - Preprocess all the FITs data, then pivot
    - Save all of the processed data to a single HDF5 database

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """
    # Detect types in data not listed in sntypes and assign as contaminant.
    # This must happen before build_traintestval_splits so that column names
    # (target_Nclasses, dataset_*_Nclasses) are computed correctly.
    detect_contaminant_types(settings)

    # Clean up data folders
    if settings.overwrite is True:
        for folder in [settings.preprocessed_dir, settings.processed_dir]:
            # Dont throw error if folder exists with exist_ok Flag.
            for f in glob.glob(f"{folder}/*"):
                os.remove(f)
            # Save cli args
            settings._save_to_json(settings.processed_dir)

    # split dataset in train test and validation
    build_traintestval_splits(settings)

    # Preprocess dataset
    preprocess_data(settings)

    # Pivot dataframe
    list_files = natsorted(glob.glob(f"{settings.preprocessed_dir}/*PHOT*"))
    pivot_dataframe_batch(list_files, settings)

    # Aggregate the pivoted dataframe
    list_files = natsorted(
        glob.glob(os.path.join(settings.preprocessed_dir, "*pivot.pickle*"))
    )
    logging_utils.print_green("Concatenating pivot")

    df = pd.concat([pd.read_pickle(f) for f in list_files], axis=0)
    # Save to HDF5
    data_utils.save_to_HDF5(settings, df)

    # Save plots to visualize the distribution of some of the data features
    try:
        SNinfo_df = data_utils.load_HDF5_SNinfo(settings)
        datasets_plots(SNinfo_df, settings)
    except Exception:
        logging_utils.print_yellow(
            "Warning: can't do data plots if no saltfit for this dataset"
        )

    # Clean preprocessed directory
    if settings.debug:
        logging_utils.print_red("Debugging mode, keeping preprocessed data")
    else:
        shutil.rmtree(settings.preprocessed_dir)

    logging_utils.print_green("Finished making dataset")



# class BayesRNNBase and BayesLSTM are developed based on RNNBase and LSTM from pytorch package (v2.0.1) (https://github.com/pytorch/pytorch/tree/v2.0.1)
# Source: https://github.com/pytorch/pytorch/blob/v2.0.1/torch/nn/modules/rnn.py

import math
import numbers
import warnings
from typing import Tuple, Optional

import torch
from torch import Tensor, _VF
from torch.nn import Parameter, Module
from torch.nn.utils.rnn import PackedSequence, pad_packed_sequence

import torch.nn.functional as F


# mean_field_inference: instead of sampling a weight as in W = N(mu, sigma)
# we always set W = mu


class BayesianRNN(Module):
    def __init__(self, input_size, settings):
        super().__init__()

        # Params
        self.layer_type = settings.layer_type
        self.hidden_size = settings.hidden_dim
        self.output_size = settings.nb_classes
        self.num_layers = settings.num_layers
        self.bidirectional = settings.bidirectional
        self.use_cuda = settings.use_cuda
        self.rnn_output_option = settings.rnn_output_option

        self.prior = Prior(settings.pi, settings.log_sigma1, settings.log_sigma2)
        self.prior_output = Prior(
            settings.pi, settings.log_sigma1_output, settings.log_sigma2_output
        )

        bidirectional_factor = 2 if self.bidirectional is True else 1
        last_input_size = (
            self.hidden_size * bidirectional_factor
            if self.rnn_output_option == "mean"
            else self.hidden_size * bidirectional_factor * self.num_layers
        )

        # Layers / nn objects
        if self.layer_type == "lstm":
            print("starting rnn layer ...")
            self.rnn_layer = BayesLSTM(
                input_size,
                self.hidden_size,
                self.prior,
                num_layers=self.num_layers,
                bidirectional=self.bidirectional,
                mu_lower=-0.05,
                mu_upper=0.05,
                rho_lower=math.log(
                    math.exp(self.prior.sigma_mix / settings.rho_scale_lower) - 1.0
                ),
                rho_upper=math.log(
                    math.exp(self.prior.sigma_mix / settings.rho_scale_upper) - 1.0
                ),
            )
            print("type of rnn layer: ", type(self.rnn_layer))
            print("finish rnn layer...")
        else:
            raise ValueError(
                "Unregistered BayesRNN mode: {}".format(self.layer_type.upper())
            )

        self.output_layer = BayesLinear(
            last_input_size,
            self.output_size,
            self.prior_output,
            mu_lower=-0.05,
            mu_upper=0.05,
            rho_lower=math.log(
                math.exp(self.prior_output.sigma_mix / settings.rho_scale_lower_output)
                - 1.0
            ),
            rho_upper=math.log(
                math.exp(self.prior_output.sigma_mix / settings.rho_scale_upper_output)
                - 1.0
            ),
        )

    def forward(self, x, mean_field_inference=False):
        # out = packed output from last layer
        # out has dim (seq_len, batch_size, hidden_size) when unpacked
        # hidden = (hn, cn) for lstm
        # hidden = hn for GRU and RNN
        # hn has dim (num_layers * num_directions, batch, hidden_size)
        # cn has dim (num_layers * num_directions, batch, hidden_size)
        # assuming num_directions = 1, num_layers = 2 :
        # hn[-1, -1] == out[len, -1] where len is the len of the seq at batch index == -1
        x, hidden = self.rnn_layer(x, mean_field_inference=mean_field_inference)
        # Output options
        # Standard: all layers, only end of pass
        #    - take last pass in all layers (hidden)
        #    - reshape and apply dropout
        #    - use h20 to obtain output (h2o input: hidden_size*num_layers*bi)
        # Mean: last layer, mean on sequence
        #    - take packed output from last layer (out) that contains all time steps for the last layer
        #    - find where padding was done and create a mask for those values, apply this mask
        #    - take a mean for the whole sequence (time_steps)
        #    - use h2o to obtain output (beware! it is only one layer deep since it is the last one only)
        if self.rnn_output_option == "standard":
            # Special case for lstm where hidden = (h, c)
            if self.layer_type == "lstm":
                hn = hidden[0]
            else:
                hn = hidden

            hn = hn.permute(1, 2, 0).contiguous()
            # hn is (num_layers * num_directions, batch, hidden_size)
            batch_size = hn.shape[0]
            # hn now is (batch, hidden size, num_layers * num_directions)
            x = hn.view(batch_size, -1)
            # x is (batch, hidden size * num_layers * num_directions)

        if self.rnn_output_option == "mean":
            if isinstance(x, PackedSequence):
                x, lens = pad_packed_sequence(x)
                # x is (seq_len, batch, hidden size)

                # take mean over seq_len
                x = x.sum(0) / lens.unsqueeze(-1).float().to(x.device)
                # x is (batch, hidden_size)
            else:
                x = x.mean(0)
                # x is (batch, hidden_size)

        # Final projection layer
        output = self.output_layer(x, mean_field_inference=mean_field_inference)

        # Compute KL
        self.kl = self.rnn_layer.kl + self.output_layer.kl

        return output


def _apply_permutation(tensor: Tensor, permutation: Tensor, dim: int = 1) -> Tensor:
    return tensor.index_select(dim, permutation)


class BayesRNNBase(Module):
    def __init__(
        self,
        mode: str,
        input_size: int,
        hidden_size: int,
        prior: object,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = False,
        dropout: float = 0,
        bidirectional: bool = False,
        mu_lower=-0.05,
        mu_upper=0.05,
        rho_lower=-1,
        rho_upper=-4,
    ):
        super().__init__()
        self.mode = mode
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.prior = prior
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.dropout_state = {}
        self.bidirectional = bidirectional
        num_directions = 2 if bidirectional else 1

        if (
            not isinstance(dropout, numbers.Number)
            or not 0 <= dropout <= 1
            or isinstance(dropout, bool)
        ):
            raise ValueError(
                "dropout should be a number in range [0, 1] "
                "representing the probability of an element being "
                "zeroed"
            )
        if dropout > 0 and num_layers == 1:
            warnings.warn(
                "dropout option adds dropout after all but last "
                "recurrent layer, so non-zero dropout expects "
                "num_layers greater than 1, but got dropout={} and "
                "num_layers={}".format(dropout, num_layers)
            )

        if mode == "LSTM":
            gate_size = 4 * hidden_size
        elif mode == "GRU":
            gate_size = 3 * hidden_size
        else:
            gate_size = hidden_size

        self._flat_weights_names = []
        self._all_weights = []
        for layer in range(num_layers):
            for direction in range(num_directions):
                layer_input_size = (
                    input_size if layer == 0 else hidden_size * num_directions
                )

                w_ih_mu, w_ih_rho = get_bbb_variable(
                    (gate_size, layer_input_size),
                    mu_lower,
                    mu_upper,
                    rho_lower,
                    rho_upper,
                )

                w_hh_mu, w_hh_rho = get_bbb_variable(
                    (gate_size, hidden_size), mu_lower, mu_upper, rho_lower, rho_upper
                )

                b_ih_mu, b_ih_rho = get_bbb_variable(
                    (gate_size,), mu_lower, mu_upper, rho_lower, rho_upper
                )
                b_hh_mu, b_hh_rho = get_bbb_variable(
                    (gate_size,), mu_lower, mu_upper, rho_lower, rho_upper
                )

                if bias:
                    layer_params = (
                        w_ih_mu,
                        w_ih_rho,
                        w_hh_mu,
                        w_hh_rho,
                        b_ih_mu,
                        b_ih_rho,
                        b_hh_mu,
                        b_hh_rho,
                    )
                else:
                    layer_params = (
                        w_ih_mu,
                        w_ih_rho,
                        w_hh_mu,
                        w_hh_rho,
                    )

                suffix = "_reverse" if direction == 1 else ""
                param_names = [
                    "weight_ih_mu_l{}{}",
                    "weight_ih_rho_l{}{}",
                    "weight_hh_mu_l{}{}",
                    "weight_hh_rho_l{}{}",
                ]

                if bias:
                    param_names += [
                        "bias_ih_mu_l{}{}",
                        "bias_ih_rho_l{}{}",
                        "bias_hh_mu_l{}{}",
                        "bias_hh_rho_l{}{}",
                    ]
                param_names = [x.format(layer, suffix) for x in param_names]

                for name, param in zip(param_names, layer_params):
                    setattr(self, name, param)
                self._flat_weights_names.extend(param_names)
                self._all_weights.append(param_names)

        self.flatten_parameters()
        self.kl = None

    def flatten_parameters(self) -> None:
        """Resets parameter data pointer so that they can use faster code paths."""
        self._data_ptrs = []
        self._param_buf_size = 0
        return

    def _apply(self, fn):
        ret = super(BayesRNNBase, self)._apply(fn)
        self.flatten_parameters()
        return ret

    def check_input(self, input: Tensor, batch_sizes: Optional[Tensor]) -> None:
        expected_input_dim = 2 if batch_sizes is not None else 3
        if input.dim() != expected_input_dim:
            raise RuntimeError(
                "input must have {} dimensions, got {}".format(
                    expected_input_dim, input.dim()
                )
            )
        if self.input_size != input.size(-1):
            raise RuntimeError(
                "input.size(-1) must be equal to input_size. Expected {}, got {}".format(
                    self.input_size, input.size(-1)
                )
            )

    def get_expected_hidden_size(
        self, input: Tensor, batch_sizes: Optional[Tensor]
    ) -> Tuple[int, int, int]:
        if batch_sizes is not None:
            mini_batch = int(batch_sizes[0])
        else:
            mini_batch = input.size(0) if self.batch_first else input.size(1)
        num_directions = 2 if self.bidirectional else 1

        expected_hidden_size = (
            self.num_layers * num_directions,
            mini_batch,
            self.hidden_size,
        )

        return expected_hidden_size

    def check_hidden_size(
        self,
        hx: Tensor,
        expected_hidden_size: Tuple[int, int, int],
        msg: str = "Expected hidden size {}, got {}",
    ) -> None:
        if hx.size() != expected_hidden_size:
            raise RuntimeError(msg.format(expected_hidden_size, list(hx.size())))

    def check_forward_args(
        self, input: Tensor, hidden: Tensor, batch_sizes: Optional[Tensor]
    ):
        self.check_input(input, batch_sizes)
        expected_hidden_size = self.get_expected_hidden_size(input, batch_sizes)

        self.check_hidden_size(hidden, expected_hidden_size)

    def permute_hidden(self, hx: Tensor, permutation: Optional[Tensor]):
        if permutation is None:
            return hx
        return _apply_permutation(hx, permutation)

    def extra_repr(self) -> str:
        s = "{input_size}, {hidden_size}"
        if self.num_layers != 1:
            s += ", num_layers={num_layers}"
        if self.bias is not True:
            s += ", bias={bias}"
        if self.batch_first is not False:
            s += ", batch_first={batch_first}"
        if self.dropout != 0:
            s += ", dropout={dropout}"
        if self.bidirectional is not False:
            s += ", bidirectional={bidirectional}"
        return s.format(**self.__dict__)


class BayesLSTM(BayesRNNBase):
    def __init__(self, *args, **kwargs):
        super().__init__("LSTM", *args, **kwargs)

    def get_expected_cell_size(
        self, input: Tensor, batch_sizes: Optional[Tensor]
    ) -> Tuple[int, int, int]:
        if batch_sizes is not None:
            mini_batch = int(batch_sizes[0])
        else:
            mini_batch = input.size(0) if self.batch_first else input.size(1)
        num_directions = 2 if self.bidirectional else 1
        expected_hidden_size = (
            self.num_layers * num_directions,
            mini_batch,
            self.hidden_size,
        )
        return expected_hidden_size

    def check_forward_args(
        self,  # type: ignore[override]
        input: Tensor,
        hidden: Tuple[Tensor, Tensor],
        batch_sizes: Optional[Tensor],
    ):
        self.check_input(input, batch_sizes)
        self.check_hidden_size(
            hidden[0],
            self.get_expected_hidden_size(input, batch_sizes),
            "Expected hidden[0] size {}, got {}",
        )
        self.check_hidden_size(
            hidden[1],
            self.get_expected_cell_size(input, batch_sizes),
            "Expected hidden[1] size {}, got {}",
        )

    def permute_hidden(
        self,  # type: ignore[override]
        hx: Tuple[Tensor, Tensor],
        permutation: Optional[Tensor],
    ) -> Tuple[Tensor, Tensor]:
        if permutation is None:
            return hx
        return _apply_permutation(hx[0], permutation), _apply_permutation(
            hx[1], permutation
        )

    def forward(self, input, hx=None, mean_field_inference=False):
        is_packed = isinstance(input, PackedSequence)
        if is_packed:
            input, batch_sizes, sorted_indices, unsorted_indices = input
            max_batch_size = batch_sizes[0]
            max_batch_size = int(max_batch_size)
        else:
            batch_sizes = None
            assert input.dim() in (
                2,
                3,
            ), f"LSTM: Expected input to be 2-D or 3-D but received {input.dim()}-D tensor"
            is_batched = input.dim() == 3
            batch_dim = 0 if self.batch_first else 1
            if not is_batched:
                input = input.unsqueeze(batch_dim)
            max_batch_size = input.size(0) if self.batch_first else input.size(1)
            sorted_indices = None
            unsorted_indices = None

        if hx is None:
            num_directions = 2 if self.bidirectional else 1

            hx = torch.zeros(
                self.num_layers * num_directions,
                max_batch_size,
                self.hidden_size,
                dtype=input.dtype,
                device=input.device,
            )

            hx = (hx, hx)

        # else:
        #     if batch_sizes is None:  # If not PackedSequence input.
        #         if is_batched:
        #             if (hx[0].dim() != 3 or hx[1].dim() != 3):
        #                 msg = ("For batched 3-D input, hx and cx should "
        #                         f"also be 3-D but got ({hx[0].dim()}-D, {hx[1].dim()}-D) tensors")
        #                 raise RuntimeError(msg)
        #         else:
        #             if hx[0].dim() != 2 or hx[1].dim() != 2:
        #                 msg = ("For unbatched 2-D input, hx and cx should "
        #                         f"also be 2-D but got ({hx[0].dim()}-D, {hx[1].dim()}-D) tensors")
        #                 raise RuntimeError(msg)
        #             hx = (hx[0].unsqueeze(1), hx[1].unsqueeze(1))

        # # Each batch of the hidden state should match the input sequence that
        # # the user believes he/she is passing in.
        # hx = self.permute_hidden(hx, sorted_indices)

        # todo: memory usage efficiency
        # has_flat_weights = (
        #     list(p.data.data_ptr() for p in self.parameters()) == self._data_ptrs
        # )
        # if has_flat_weights:
        #     first_data = next(self.parameters()).data
        #     assert first_data.storage().size() == self._param_buf_size
        #     flat_weight = first_data.new().set_(
        #         first_data.storage(), 0, torch.Size([self._param_buf_size])
        #     )
        # else:
        #     flat_weight = None

        self.check_forward_args(input, hx, batch_sizes)

        # Format weights for BBB
        all_weights = []
        num_layers = self.num_layers
        num_directions = 2 if self.bidirectional else 1
        self.kl = 0
        # Loop over layers
        for layer_idx in range(num_layers):
            # Loop over directions
            for direction in range(num_directions):

                suffix = "_reverse" if direction == 1 else ""

                w_ih_mean = getattr(self, f"weight_ih_mu_l{layer_idx}{suffix}")
                w_ih_rho = getattr(self, f"weight_ih_rho_l{layer_idx}{suffix}")
                w_ih_sigma = F.softplus(w_ih_rho) + 1e-5

                w_hh_mean = getattr(self, f"weight_hh_mu_l{layer_idx}{suffix}")
                w_hh_rho = getattr(self, f"weight_hh_rho_l{layer_idx}{suffix}")
                w_hh_sigma = F.softplus(w_hh_rho) + 1e-5

                if mean_field_inference:
                    weight_ih = w_ih_mean
                    weight_hh = w_hh_mean

                else:
                    # Sample weights from normal distribution
                    eps_ih = w_ih_mean.data.new(w_ih_mean.size()).normal_(0.0, 1.0)
                    weight_ih = w_ih_mean + eps_ih * w_ih_sigma

                    eps_hh = w_hh_mean.data.new(w_hh_mean.size()).normal_(0.0, 1.0)
                    weight_hh = w_hh_mean + eps_hh * w_hh_sigma

                all_weights.extend([weight_ih, weight_hh])

                # Compute KL divergence
                self.kl += compute_KL(weight_ih, w_ih_mean, w_ih_sigma, self.prior)
                self.kl += compute_KL(weight_hh, w_hh_mean, w_hh_sigma, self.prior)

                if self.bias:

                    b_ih_mean = getattr(self, f"bias_ih_mu_l{layer_idx}{suffix}")
                    b_ih_rho = getattr(self, f"bias_ih_rho_l{layer_idx}{suffix}")
                    b_ih_sigma = F.softplus(b_ih_rho) + 1e-5

                    b_hh_mean = getattr(self, f"bias_hh_mu_l{layer_idx}{suffix}")
                    b_hh_rho = getattr(self, f"bias_hh_rho_l{layer_idx}{suffix}")
                    b_hh_sigma = F.softplus(b_hh_rho) + 1e-5

                    if mean_field_inference:
                        bias_ih = b_ih_mean
                        bias_hh = b_hh_mean

                    else:
                        eps_ih = b_ih_mean.data.new(b_ih_mean.size()).normal_(0.0, 1.0)
                        bias_ih = b_ih_mean + eps_ih * b_ih_sigma

                        eps_hh = b_hh_mean.data.new(b_hh_mean.size()).normal_(0.0, 1.0)
                        bias_hh = b_hh_mean + eps_hh * b_hh_sigma

                    all_weights.extend([bias_ih, bias_hh])

                    self.kl += compute_KL(bias_ih, b_ih_mean, b_ih_sigma, self.prior)
                    self.kl += compute_KL(bias_hh, b_hh_mean, b_hh_sigma, self.prior)

        if batch_sizes is None:
            result = _VF.lstm(
                input,
                hx,
                all_weights,
                self.bias,
                self.num_layers,
                self.dropout,
                self.training,
                self.bidirectional,
                self.batch_first,
            )
        else:
            result = _VF.lstm(
                input,
                batch_sizes,
                hx,
                all_weights,
                self.bias,
                self.num_layers,
                self.dropout,
                self.training,
                self.bidirectional,
            )
        output = result[0]
        hidden = result[1:]

        if is_packed:
            output_packed = PackedSequence(
                output, batch_sizes, sorted_indices, unsorted_indices
            )
            return output_packed, self.permute_hidden(hidden, unsorted_indices)
        else:
            if not is_batched:
                output = output.squeeze(batch_dim)
                hidden = (hidden[0].squeeze(1), hidden[1].squeeze(1))
            return output, self.permute_hidden(hidden, unsorted_indices)


class BayesGRU(BayesRNNBase):
    def __init__(self, *args, **kwargs):
        super().__init__("GRU", *args, **kwargs)


class BayesLinear(Module):
    def __init__(
        self, in_features, out_features, prior, mu_lower, mu_upper, rho_lower, rho_upper
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.prior = prior

        mu, rho = get_bbb_variable(
            (out_features, in_features), mu_lower, mu_upper, rho_lower, rho_upper
        )

        bias = Parameter(torch.Tensor(out_features))
        bias.data.fill_(0.0)

        self.mu = mu
        self.rho = rho
        self.bias = bias
        self.kl = None

    def forward(self, input, mean_field_inference=False):

        # Sample weight
        mean = self.mu
        sigma = F.softplus(self.rho) + 1e-5

        if mean_field_inference:
            weights = mean
        else:
            # Sample weights from normal distribution
            # This way of creating the epsilon variable is faster than
            # from numpy or torch.randn or FloatTensor.normal_ when mean is already
            # on the GPU
            eps = mean.data.new(mean.size()).normal_(0.0, 1.0)
            weights = mean + eps * sigma

        logits = F.linear(input, weights, self.bias)

        # Compute KL divergence
        self.kl = compute_KL(weights, mean, sigma, self.prior)

        return logits

    def __repr__(self):
        return (
            self.__class__.__name__
            + " ("
            + str(self.in_features)
            + " -> "
            + str(self.out_features)
            + ")"
        )


class BayesBiasLinear(Module):
    def __init__(
        self, in_features, out_features, prior, mu_lower, mu_upper, rho_lower, rho_upper
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.prior = prior

        self.W_mu, self.W_rho = get_bbb_variable(
            (out_features, in_features), mu_lower, mu_upper, rho_lower, rho_upper
        )

        self.b_mu, self.b_rho = get_bbb_variable(
            (out_features,), mu_lower, mu_upper, rho_lower, rho_upper
        )

    def forward(self, X, mean_field_inference=False):
        # Sample weight
        W_mean = self.W_mu
        W_sigma = F.softplus(self.W_rho) + 1e-5

        b_mean = self.b_mu
        b_sigma = F.softplus(self.b_rho) + 1e-5

        if mean_field_inference:
            weights = W_mean
            biases = b_mean
        else:
            # Sample weights from normal distribution
            # This way of creating the epsilon variable is faster than
            # from numpy or torch.randn or FloatTensor.normal_ when mean is already
            # on the GPU
            W_eps = W_mean.data.new(W_mean.size()).normal_(0.0, 1.0)
            weights = W_mean + W_eps * W_sigma

            b_eps = b_mean.data.new(b_mean.size()).normal_(0.0, 1.0)
            biases = b_mean + b_eps * b_sigma

        logits = F.linear(X, weights, biases)

        # Compute KL divergence
        self.kl = compute_KL(weights, W_mean, W_sigma, self.prior)
        self.kl += compute_KL(biases, b_mean, b_sigma, self.prior)

        return logits


class BayesEmbedding(Module):
    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        prior,
        mu_lower,
        mu_upper,
        rho_lower,
        rho_upper,
    ):
        super().__init__()

        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.prior = prior
        self.max_norm = None
        self.norm_type = 2
        self.scale_grad_by_freq = False
        self.sparse = False
        self.padding_idx = -1

        mu, rho = get_bbb_variable(
            [num_embeddings, embedding_dim], mu_lower, mu_upper, rho_lower, rho_upper
        )

        self.mu = mu
        self.rho = rho
        self.kl = None

    def forward(self, input, mean_field_inference=False):

        # Sample weight
        mean = self.mu
        sigma = F.softplus(self.rho) + 1e-5

        if mean_field_inference:
            weights = mean
        else:
            # This way of creating the epsilon variable is faster than
            # from numpy or torch.randn or FloatTensor.normal_ when mean is already
            # on the GPU
            eps = mean.data.new(mean.size()).normal_(0.0, 1.0)
            weights = mean + eps * sigma

        # Compute KL divergence
        self.kl = compute_KL(weights, mean, sigma, self.prior)

        after_embed = F.embedding(
            input,
            weights,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )

        return after_embed

    def __repr__(self):
        s = "{name}({num_embeddings}, {embedding_dim}"
        if self.padding_idx is not None:
            s += ", padding_idx={padding_idx}"
        if self.max_norm is not None:
            s += ", max_norm={max_norm}"
        if self.norm_type != 2:
            s += ", norm_type={norm_type}"
        if self.scale_grad_by_freq is not False:
            s += ", scale_grad_by_freq={scale_grad_by_freq}"
        if self.sparse is not False:
            s += ", sparse=True"
        s += ")"
        return s.format(name=self.__class__.__name__, **self.__dict__)


class Prior(object):
    def __init__(self, pi=0.25, log_sigma1=-1.0, log_sigma2=-7.0):
        self.pi_mixture = pi
        self.log_sigma1 = log_sigma1
        self.log_sigma2 = log_sigma2
        self.sigma1 = math.exp(log_sigma1)
        self.sigma2 = math.exp(log_sigma2)

        self.sigma_mix = math.sqrt(
            pi * math.pow(self.sigma1, 2) + (1.0 - pi) * math.pow(self.sigma2, 2)
        )


def get_bbb_variable(shape, mu_lower, mu_upper, rho_lower, rho_upper):

    mu = Parameter(torch.Tensor(*shape))
    rho = Parameter(torch.Tensor(*shape))

    # Initialize
    mu.data.uniform_(mu_lower, mu_upper)
    rho.data.uniform_(rho_lower, rho_upper)

    return mu, rho


def logsumexp(x, dim):
    """Logsumexp trick to avoid overflow in a log of sum of exponential expression

    Args:
        x (Tensor): the input on which to compute the log of sum of exponential

    Returns:
        logsum (Tensor): the computed log of sum of exponential
    """

    assert x.dim() == 2
    x_max, x_max_idx = x.max(dim=dim, keepdim=True)
    logsum = x_max + torch.log((x - x_max).exp().sum(dim=dim, keepdim=True))
    return logsum


def compute_KL(x, mu, sigma, prior):

    posterior = torch.distributions.Normal(mu.view(-1), sigma.view(-1))
    log_posterior = posterior.log_prob(x.view(-1)).sum()

    if x.is_cuda:
        n1 = torch.distributions.Normal(
            torch.tensor([0.0]).cuda(), torch.tensor([prior.sigma1]).cuda()
        )
        n2 = torch.distributions.Normal(
            torch.tensor([0.0]).cuda(), torch.tensor([prior.sigma2]).cuda()
        )
    else:
        n1 = torch.distributions.Normal(0.0, prior.sigma1)
        n2 = torch.distributions.Normal(0.0, prior.sigma2)

    mix1 = torch.sum(n1.log_prob(x)) + math.log(prior.pi_mixture)
    mix2 = torch.sum(n2.log_prob(x)) + math.log(1.0 - prior.pi_mixture)
    prior_mix = torch.stack([mix1, mix2])
    log_prior = torch.logsumexp(prior_mix, 0)

    return log_posterior - log_prior



import os
import sys
import json
import yaml
import argparse
from pathlib import Path
from natsort import natsorted
from collections import OrderedDict
from distutils.util import strtobool
from .utils import experiment_settings
from supernnova.utils import logging_utils as lu

# group options to show in print_help for different actions
COMMON_OPTIONS = [
    "--seed",
    "--use_cuda",
    "--dump_dir",
    "--config_file",
    "--no_dump",
    "--help",
]

MAKE_DATA_OPTIONS = COMMON_OPTIONS + [
    "--data_fraction",
    "--data_testing",
    "--data_training",
    "--debug",
    "--norm",
    "--no_overwrite",
    "--raw_dir",
    "--fits_dir",
    "--testing_ids",
    "--photo_window_files",
    "--photo_window_var",
    "--photo_window_min",
    "--photo_window_max",
    "--list_filters",
    "--phot_reject",
    "--phot_reject_list",
    "--redshift",
    "--redshift_label",
    "--explore_lightcurves",
]

TRAIN_RNN_OPTIONS = COMMON_OPTIONS + [
    "--cyclic",
    "--cyclic_phases",
    "--random_length",
    "--random_redshift",
    "--weight_decay",
    "--layer_type",
    "--model",
    "--learning_rate",
    "--nb_classes",
    "--sntypes",
    "--sntype_var",
    "--additional_train_var",
    "--nb_epoch",
    "--batch_size",
    "--hidden_dim",
    "--num_layers",
    "--dropout",
    "--bidirectional",
    "--rnn_output_option",
    "--pi",
    "--log_sigma1",
    "--log_sigma2",
    "--rho_scale_lower",
    "--rho_scale_upper",
    "--log_sigma1_output",
    "--log_sigma2_output",
    "--rho_scale_lower_output",
    "--rho_scale_upper_output",
    "--num_inference_samples",
    "--mean_field_inference",
    "--monitor_interval",
    "--calibration",
    "--plot_file",
    "--swag",
    "--swag_pretrained_path",
    "--swag_start_epoch",
    "--swag_samples",
    "--swag_scale",
    "--swag_no_lr_cov",
]

VALIDATE_RNN_OPTIONS = COMMON_OPTIONS + [
    "--model_files",
    "--plot_lcs",
    "--plot_file",
    "--calibration",
    "--plot_prediction_distribution",
    "--speed",
    "--swag",
    "--swag_finetune",
    "--swag_samples",
    "--swag_scale",
    "--swag_no_lr_cov",
]

SHOW_OPTIONS = COMMON_OPTIONS + [
    "--model_files",
    "--plot_lcs",
    "--plot_file",
    "--plot_prediction_distribution",
    "--prediction_files",
    "--calibration",
]

PERFORMANCE_OPTIONS = COMMON_OPTIONS + [
    "--metrics",
    "--prediction_files",
    "--speed",
    "--done_files",
]

helps = {
    "make_data": MAKE_DATA_OPTIONS,
    "train_rnn": TRAIN_RNN_OPTIONS,
    "validate_rnn": VALIDATE_RNN_OPTIONS,
    "show": SHOW_OPTIONS,
    "performance": PERFORMANCE_OPTIONS,
}


# customize help, so it only print out relevant to the provide command
def generate_command_help(parser, command_arg):
    """Generate a help message for specific command options."""
    help_message = f"usage: snn {command_arg} [options]\n\n"
    help_message += "optional arguments:\n"

    # Filter and sort actions based on their option strings
    command_options = helps[command_arg]
    relevant_actions = [
        action
        for action in parser._actions
        if any(opt in command_options for opt in action.option_strings)
    ]

    sorted_actions = sorted(relevant_actions, key=lambda a: a.option_strings[0])

    for action in sorted_actions:
        option_str = ", ".join(action.option_strings)
        help_descr = f"{option_str:30} {action.help}"
        help_message += f"  {help_descr}\n"
    return help_message


def handle_custom_help(parser, command_arg):
    """Display help messages for provided command"""

    if command_arg in helps.keys():
        print(generate_command_help(parser, command_arg))
    else:
        print("The command {} is not valid.".format(command_arg))
        sys.exit()


class CustomHelpAction(argparse.Action):
    command_arg = None

    def __init__(self, option_strings, dest, help=None):
        super(CustomHelpAction, self).__init__(
            option_strings=option_strings, dest=dest, nargs=0, help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        handle_custom_help(parser, self.__class__.command_arg)
        parser.exit()  # Exit after printing the custom help message


def absolute_path(path):
    return str(Path(path).resolve())


def absolute_path_or_HEAD(value):
    if value == "HEAD":
        return value
    else:
        return os.path.abspath(value)


def get_args(command_arg):

    CustomHelpAction.command_arg = command_arg

    parser = argparse.ArgumentParser(description="SNIa classification", add_help=False)

    parser.add_argument(
        "--help", action=CustomHelpAction, help="Show custom help message"
    )

    parser.add_argument("--seed", type=int, default=0, help="Random seed to be used")

    parser.add_argument(
        "--explore_lightcurves",  # use it without using debbug
        action="store_true",
        help="Plot a random selection of lightcurves",
    )
    parser.add_argument(
        "--speed", action="store_true", help="Get RNN speed benchmark"
    )  # test this option!

    parser.add_argument(
        "--monitor_interval",
        type=int,
        default=1,
        help="Monitor validation every monitor_interval epoch",
    )
    parser.add_argument(
        "--metrics",
        action="store_true",
        help="Use Pred file to compute metrics",  #  test this option!
    )

    # parser.add_argument(
    #     "--performance", # elimiate this option
    #     action="store_true",
    #     help="Get method performance and paper plots",
    # )

    parser.add_argument(
        "--calibration",
        action="store_true",
        help="Plot calibration of trained classifiers",
    )

    parser.add_argument(
        "--plot_lcs",  # test this option
        action="store_true",
        help="Plot lcs with classification probabilities",
    )

    parser.add_argument(
        "--plot_file",  # test this option
        default=None,
        help="Plot subset of lcs in file (csv with SNID column)",
    )

    parser.add_argument(
        "--plot_prediction_distribution",
        action="store_true",
        help="Plot lcs and the histogram of probability for each class",
    )
    parser.add_argument(
        "--model_files", nargs="+", type=absolute_path, help="Path to model files"
    )  # test it
    parser.add_argument(
        "--prediction_files",
        nargs="+",
        type=absolute_path,
        help="Path to prediction files",  # test it
    )

    parser.add_argument(
        "--metric_files", nargs="+", type=absolute_path, help="Path to metric files"
    )

    parser.add_argument(
        "--done_file",
        default=None,
        type=absolute_path,
        help="Done or failure file name",
    )

    parser.add_argument(
        "--no_dump", action="store_true", help="No dump database nor preds"
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Debug database creation: one file processed only",
    )

    ########################
    # Data parameters
    ########################
    # dir_path = os.path.dirname(os.path.realpath(__file__))
    # default_dump_dir = str(Path(dir_path).parent.parent / "snndump")
    default_dump_dir = absolute_path("snndump")
    parser.add_argument(
        "--dump_dir",
        type=absolute_path,
        default=default_dump_dir,
        help="Default path where data and models are dumped",
    )

    parser.add_argument(
        "--fits_dir",
        type=absolute_path,
        default=f"{default_dump_dir}/fits",
        help="Default path where fits to photometry are",
    )

    parser.add_argument(
        "--raw_dir",
        type=absolute_path,
        default=f"{default_dump_dir}/raw",
        help="Default path where raw data is",
    )

    parser.add_argument(
        "--redshift",  # change it by Anais
        choices=["none", "zpho", "zspe"],
        default="none",
        help="Host redshift used in classification: none, zpho, zspe",
    )

    parser.add_argument(
        "--norm",
        choices=["none", "perfilter", "global", "cosmo", "cosmo_quantile"],
        default="global",
        help="Feature normalization: global does the same norm for all filters",
    )

    parser.add_argument(
        "--source_data",
        choices=["saltfit", "photometry"],
        default="photometry",
        help="Data source used to select light-curves for supernnova",
    )

    parser.add_argument(
        "--no_overwrite",
        action="store_true",
        help="If True: do not clean processed_dir and preprocessed_dir when calling `python run.py --data`",
    )

    parser.add_argument(
        "--data_fraction", type=float, default=1.0, help="Fraction of data to use"
    )

    parser.add_argument(
        "--data_training",
        action="store_true",
        help="Create database with mostly training set of 99.5%%",
    )

    parser.add_argument(
        "--data_testing",
        action="store_true",
        help="Create database with only validation set",
    )

    parser.add_argument(
        "--testing_ids",  # test it
        default=None,
        help="Filename with SNIDs to be used for testing (.csv with SNID column or .npy)",
    )

    # Photometry window
    parser.add_argument(
        "--photo_window_files",
        nargs="+",
        type=absolute_path_or_HEAD,
        help="Path(s) to fits files with PEAKMJD estimation or 'HEAD'",
    )

    parser.add_argument(
        "--photo_window_var",
        type=str,
        default="PKMJDINI",
        help="Variable representing PEAKMJD for photo window (in photo_window_files)",
    )

    parser.add_argument(
        "--photo_window_min", type=int, default=-30, help="Window size before peak"
    )

    parser.add_argument(
        "--photo_window_max", type=int, default=100, help="Window size after peak"
    )

    # Survey configuration
    parser.add_argument(
        "--list_filters",  # test it
        nargs="+",
        default=natsorted(["g", "i", "r", "z"]),
        help="Survey filters",
    )

    # Photometry filtering
    parser.add_argument(
        "--phot_reject",
        type=None,
        help="Variable for photometry flag rejection as a power of 2 (e.g.PHOTFLAG)",
    )

    parser.add_argument(
        "--phot_reject_list",
        nargs="+",
        type=int,
        default=[8, 16, 32, 64, 128, 256, 512],
        help="Bit list to mask (supports only powers of 2)",
    )

    parser.add_argument(
        "--redshift_label",
        type=str,
        default="none",
        help="Redshift label to be used instead of HOSTGAL_SPECZ (with _ERR) and SIM_REDSHIFT",
    )

    ######################
    # RNN  parameters
    ######################
    parser.add_argument(
        "--cyclic", action="store_true", help="Use cyclic learning rate"
    )

    parser.add_argument(
        "--cyclic_phases", nargs=3, default=[5, 10, 15], type=int, help="Cyclic phases"
    )

    parser.add_argument(
        "--random_length",
        choices=[True, False],
        default=True,
        type=lambda x: bool(strtobool(x)),
        help="Use random length sequences for training",
    )

    parser.add_argument(
        "--random_redshift",
        action="store_true",
        help="randomly set spectroscopic redshift to -1 (i.e. unknown)",
    )

    parser.add_argument(
        "--weight_decay",
        type=float,
        default=0.0000001,
        help="L2 decay on weights (variational)",
    )

    parser.add_argument(
        "--layer_type",
        default="lstm",
        type=str,
        choices=["lstm", "gru", "rnn"],
        help="recurrent layer type",
    )

    parser.add_argument(
        "--model",
        default="vanilla",  # Anais change the name
        type=str,
        choices=["vanilla", "variational", "bayesian", "bayesian_2"],
        help="recurrent model type",
    )

    parser.add_argument(
        "--use_cuda", action="store_true", help="Use GPU (pytorch backend only)"  # test
    )

    parser.add_argument(
        "--learning_rate", default=1e-3, type=float, help="Learning rate"
    )

    parser.add_argument(
        "--nb_classes",
        default=2,
        type=int,
        help="Number of classification targets",  # test
    )

    parser.add_argument(
        "--sntypes",  # test it
        default=OrderedDict(
            {
                "101": "Ia",
                "120": "IIP",
                "121": "IIn",
                "122": "IIL1",
                "123": "IIL2",
                "132": "Ib",
                "133": "Ic",
            }
        ),
        type=json.loads,
        help="SN classes in sims",
    )

    parser.add_argument(
        "--target_sntype",
        type=str,
        default="Ia",
        help="Class value in --sntypes to use as target 0 (class of interest) "
        "for binary classification. Default: 'Ia'",
    )

    parser.add_argument(
        "--sntype_var",
        type=str,
        default="SNTYPE",
        help="Variable representing event types (e.g. SNTYPE)",
    )

    parser.add_argument(
        "--additional_train_var",  # test
        nargs="+",
        help="Additional training variables",
    )

    parser.add_argument(
        "--nb_epoch", default=90, type=int, help="Number of batches per epoch"
    )

    parser.add_argument("--batch_size", default=128, type=int, help="Batch size")

    parser.add_argument(
        "--hidden_dim", default=32, type=int, help="Hidden layer dimension"
    )

    parser.add_argument(
        "--num_layers", default=2, type=int, help="Number of recurrent layers"
    )

    parser.add_argument("--dropout", default=0.05, type=float, help="Dropout value")

    parser.add_argument(
        "--bidirectional",
        choices=[True, False],
        default=True,
        type=lambda x: bool(strtobool(x)),
        help="Use bidirectional models",
    )

    parser.add_argument(
        "--rnn_output_option",
        default="mean",
        type=str,
        choices=["standard", "mean"],
        help="RNN output options",
    )

    parser.add_argument("--pi", default=0.75, type=float)

    ### change to pytorch higher version
    parser.add_argument("--log_sigma1", default=-1.0, type=float)
    parser.add_argument("--log_sigma2", default=-7.0, type=float)
    parser.add_argument("--rho_scale_lower", default=4.0, type=float)
    parser.add_argument("--rho_scale_upper", default=3.0, type=float)

    # Different parameters for output layer to obtain better uncertainty
    parser.add_argument("--log_sigma1_output", default=-1.0, type=float)
    parser.add_argument("--log_sigma2_output", default=-7.0, type=float)
    parser.add_argument("--rho_scale_lower_output", default=4.0, type=float)
    parser.add_argument("--rho_scale_upper_output", default=3.0, type=float)
    ### pytorch higher version end

    parser.add_argument(
        "--num_inference_samples",
        type=int,
        default=50,
        help="Number of samples to use for Bayesian inference",
    )

    parser.add_argument(
        "--mean_field_inference",
        action="store_true",
        help="Use mean field inference for bayesian models",
    )

    parser.add_argument(
        "--config_file", default=None, type=absolute_path, help="YML config file"
    )

    parser.add_argument(
        "--swag",
        action="store_true",
        help="enable SWAG",
    )
    parser.add_argument(
        "--swag_pretrained_path",
        type=str,
        help="Run SWAG on a pretrained model",
    )
    parser.add_argument(
        "--swag_start_epoch",
        type=int,
        default=83,
        help="Epoch from which SWAG averaging begins",
    )

    parser.add_argument(
        "--swag_samples",
        type=int,
        default=30,
        help="The number of SWAG samples to draw",
    )

    parser.add_argument(
        "--swag_scale",
        type=float,
        default=0.5,
        help="Scale parameter for covariance; if equals to 0, SWA model is sampled",
    )

    parser.add_argument(
        "--swag_no_lr_cov",
        action="store_true",
        help="Disable calculating low-rank covariance",
    )

    args = parser.parse_args()

    # The following block of code deal with the case when YAML config file is provided alone
    # or together with other options
    if args.config_file:
        # keep track of user provide arguments
        _args = sys.argv[1:]
        _user_namespace, _ = parser._parse_known_args(
            _args, namespace=argparse.Namespace()
        )
        _user_args = vars(_user_namespace)

        # update args from config_file
        yml_args = load_config_file(args.config_file)
        for key in yml_args.keys():
            if hasattr(args, key):
                setattr(args, key, yml_args[key])
                # print(getattr(args, key))
            else:
                print("{} is not a valid option.".format(key))

        # update user provided value again in case user provides default value
        if len(_user_args) > 1:
            for key in _user_args.keys():
                setattr(args, key, _user_args[key])

    return args


def get_settings(command_arg, args=None):

    if not args:
        args = get_args(command_arg)

    # Initialize a settings instance
    settings = experiment_settings.ExperimentSettings(args, action=command_arg)
    assert args.rho_scale_lower >= args.rho_scale_upper
    return settings


def get_settings_from_dump(settings, model_or_pred_or_metrics_file):
    # Model settings
    model_dir = Path(model_or_pred_or_metrics_file).parent
    cli_file = model_dir / "cli_args.json"
    with open(cli_file, "r") as f:
        cli_args = json.load(f)

        # Unset general arguments
        for arg in [
            "explore_lightcurves",
            "dryrun",
            "metrics",
            # "performance",
            "calibration",
            "plot_lcs",
            "prediction_files",
            "config_file",
        ]:
            cli_args[arg] = False

    # Using dump/fits/raw dir from settings instead of model

    # raw and fits shouldnt change a thing
    cli_args["raw_dir"] = settings.raw_dir
    cli_args["fits_dir"] = settings.fits_dir
    cli_args["dump_dir"] = settings.dump_dir

    # and device
    cli_args["use_cuda"] = settings.use_cuda
    cli_args["device"] = settings.device

    # model files
    cli_args["model_files"] = settings.model_files

    # model files
    cli_args["plot_file"] = settings.plot_file

    # swag options
    cli_args["swag_samples"] = settings.swag_samples
    cli_args["swag_scale"] = settings.swag_scale
    cli_args["swag_no_lr_cov"] = settings.swag_no_lr_cov

    # Backward compatibility
    keys_not_in_model_settings = [
        k for k in settings.cli_args.keys() if k not in cli_args.keys()
    ]
    for k in keys_not_in_model_settings:
        cli_args[k] = settings.cli_args[k]

    # Warning for redshift
    if cli_args["redshift"] != settings.redshift:
        lu.print_red("Model forces redshift to be set as", cli_args["redshift"])

    settings = experiment_settings.ExperimentSettings(cli_args)

    # load normalization from json dump
    settings = get_norm_from_model(model_or_pred_or_metrics_file, settings)

    return settings


def get_norm_from_model(model_file, settings):

    import json
    import numpy as np

    norm_file = Path(model_file).parent / "data_norm.json"
    with open(norm_file, "r") as f:
        norm_args = json.load(f)
        list_norm = []
        for f in settings.training_features_to_normalize:
            minv = norm_args[f]["min"]
            meanv = norm_args[f]["mean"]
            stdv = norm_args[f]["std"]
            list_norm.append([minv, meanv, stdv])
    settings.arr_norm = np.array(list_norm)

    return settings


def load_config_file(config_file):
    with open(config_file, "r") as f:
        if config_file.endswith(".yml"):
            config = yaml.safe_load(f)
        if config_file.endswith(".json"):
            config = json.load(f)
    return config



import torch


class VanillaRNN(torch.nn.Module):
    def __init__(self, input_size, settings):
        super(VanillaRNN, self).__init__()

        # Params
        self.layer_type = settings.layer_type
        self.output_size = settings.nb_classes
        self.hidden_size = settings.hidden_dim
        self.num_layers = settings.num_layers
        self.dropout = settings.dropout
        self.bidirectional = settings.bidirectional
        self.use_cuda = settings.use_cuda
        self.rnn_output_option = settings.rnn_output_option

        bidirectional_factor = 2 if self.bidirectional is True else 1
        last_input_size = (
            self.hidden_size * bidirectional_factor
            if self.rnn_output_option == "mean"
            else self.hidden_size * bidirectional_factor * self.num_layers
        )

        # Define layers
        self.rnn_layer = getattr(torch.nn, self.layer_type.upper())(
            input_size,
            self.hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            bidirectional=self.bidirectional,
        )
        self.output_dropout_layer = torch.nn.Dropout(self.dropout)
        self.output_layer = torch.nn.Linear(last_input_size, self.output_size)

    def forward(self, x, mean_field_inference=False):
        # Reminder
        # out = packed output from last layer
        # out has dim (seq_len, batch_size, hidden_size) when unpacked
        # hidden = (hn, cn) for lstm (only final h from each pass and layer)
        # hidden = hn for GRU and RNN (only final h from each pass and layer)
        # hn has dim (num_layers * num_directions, batch, hidden_size)
        # cn has dim (num_layers * num_directions, batch, hidden_size)
        # assuming num_directions = 1, num_layers = 2 :
        # hn[-1, -1] == out[len, -1] where len is the len of the seq at batch index == -1

        x, hidden = self.rnn_layer(x)

        # Output options
        # Standard: all layers, only end of pass
        #    - take last pass in all layers (hidden)
        #    - reshape and apply dropout
        #    - use h20 to obtain output (h2o input: hidden_size*num_layers*bi)
        # Mean: last layer, mean on sequence
        #    - take packed output from last layer (out) that contains all time steps for the last layer
        #    - find where padding was done and create a mask for those values, apply this mask
        #    - take a mean for the whole sequence (time_steps)
        #    - use h2o to obtain output (beware! it is only one layer deep since it is the last one only)

        if self.rnn_output_option == "standard":
            # Special case for lstm where hidden = (h, c)
            if self.layer_type == "lstm":
                hn = hidden[0]
            else:
                hn = hidden

            hn = hn.permute(1, 2, 0).contiguous()
            # hn is (num_layers * num_directions, batch, hidden_size)
            batch_size = hn.shape[0]
            # hn now is (batch, hidden size, num_layers * num_directions)
            x = hn.view(batch_size, -1)
            # x is (batch, hidden size * num_layers * num_directions)

        if self.rnn_output_option == "mean":
            if isinstance(x, torch.nn.utils.rnn.PackedSequence):
                x, lens = torch.nn.utils.rnn.pad_packed_sequence(x)
                # x is (seq_len, batch, hidden size)

                # take mean over seq_len
                x = x.sum(0) / lens.unsqueeze(-1).float().to(x.device)
                # x is (batch, hidden_size)
            else:
                x = x.mean(0)

        # apply dropout
        x = self.output_dropout_layer(x)
        # Final projection layer
        output = self.output_layer(x)

        return output



import re
import numpy as np
import pandas as pd
from sklearn import metrics


def performance_metrics(df, sample_target=0):
    """Get performance metrics
    AUC: only valid for binomial classification, input proba of highest label class.

    Args:
        df (pandas.DataFrame) (str): with columns [target, predicted_target, class1]
        (optional) sample_target (str): for SNIa sample default is target 0

    Returns:
        accuracy, auc, purity, efficiency,truepositivefraction
    """
    n_targets = len(np.unique(df["target"]))

    # Accuracy & AUC
    accuracy = metrics.balanced_accuracy_score(
        df["target"].astype(int), df["predicted_target"]
    )
    accuracy = round(accuracy * 100, 2)
    if n_targets == 2:  # valid for biclass only
        auc = round(metrics.roc_auc_score(df["target"], df["class1"]), 4)
    else:
        auc = 0.0

    SNe_Ia = df[df["target"] == sample_target]
    SNe_CC = df[df["target"] != sample_target]
    TP = len(SNe_Ia[SNe_Ia["predicted_target"] == sample_target])
    FP = len(SNe_CC[SNe_CC["predicted_target"] == sample_target])

    P = len(SNe_Ia)
    N = len(SNe_CC)

    truepositivefraction = P / (P + N)
    purity = round(100 * TP / (TP + FP), 2) if (TP + FP) > 0 else 0
    efficiency = round(100.0 * TP / P, 2) if P > 0 else 0

    return accuracy, auc, purity, efficiency, truepositivefraction


def contamination_by_SNTYPE(df, settings, sample_target=0):
    """Get contamination contribution by each SN type in sample percentage

    Args:
        df (pandas.DataFrame) (str): with columns [target, predicted_target, SNTYPE]
        (optional) sample_target (str): for SNIa sample default is target 0

    Returns:
        df (pandas.DataFrame) (str): with columns [SNTYPE,contamination_percentage]
    """

    # Get contamination SNe
    df_cont = df[
        (df["target"] != sample_target) & (df["predicted_target"] == sample_target)
    ]
    sample_size = len(df[df["predicted_target"] == sample_target])
    # Get contamination percentage
    contribution_arr = []
    type_arr = []
    for typ in [int(t) for t in settings.sntypes.keys() if t != 101]:
        df_selection = df_cont[df_cont[settings.sntype_var] == typ]
        if sample_size > 1:
            contribution_arr.append(round(100 * len(df_selection) / sample_size, 2))
        else:
            contribution_arr.append(0)
        type_arr.append(typ)

    # Save contamination into df
    df_contamination_by_type = pd.DataFrame()
    df_contamination_by_type[settings.sntype_var] = np.array(type_arr)
    df_contamination_by_type["contamination_percentage"] = np.array(contribution_arr)

    return df_contamination_by_type


def get_quantity_vs_variable(
    quantity,
    variable,
    df,
    settings,
    contamination_by=None,
    nbins=10,
    intervals=False,
    mean_bins=False,
):
    """Get contamination/purity vs redshift

    Args:
        quantity (str): quantity to compute
        variable (str): variable for binning quantity
        df (pandas.DataFrame) (str):
            columns [SIM_REDSHIFT_CMB, target, predicted_target, class1/0(fraction of positives),SNTYPE]
        settings (str): superNNova settings
        contamination_by (str): if a contamination by a SNTYPE SNe is required
        nbins (str): number of bins to use for varaible binning
        intervals (str): instead of returning bin center, return right bound
        mean_bins (str): instead of returning bin center, return variable mean in bin

    Returns:
        bin center  (numpy array) (str): or right bound (interval option), variable mean in that bin (mean_bins option)
        quantity (numpy array)
    """

    # Slice in 10 redshift bins
    if "class" in variable:
        bins = np.arange(0, 11) / 10
        df["bin"], sliced_bins = pd.cut(df[variable], bins, retbins=True)
    else:
        df["bin"], sliced_bins = pd.cut(df[variable], nbins, retbins=True)

    sorted_df = df.sort_values(variable)
    bin_list = sorted_df["bin"].unique()

    quantity_list = []
    varbins_list = []
    for i, mybin in enumerate(bin_list):
        selected = sorted_df[sorted_df["bin"] == mybin]
        # get quantity
        if quantity == "contamination":
            cont_df = contamination_by_SNTYPE(selected, settings)
            if contamination_by:
                quant = cont_df[cont_df[settings.sntype_var] == contamination_by][
                    "contamination_percentage"
                ].values[0]
            else:
                quant = cont_df["contamination_percentage"].sum()
        else:
            d = {}
            for key, value in zip(
                ["accuracy", "auc", "purity", "efficiency", "truepositivefraction"],
                performance_metrics(selected),
            ):
                d[key] = value
            quant = d[quantity]
        quantity_list.append(quant)

        # now the bin centers/mean/interval list
        if intervals:
            varbins_list.append(mybin.right)
        elif mean_bins:
            varbins_list.append(selected[variable].mean())
        else:
            varbins_list.append(mybin.left + (mybin.right - mybin.left) / 2.0)
    return np.array(varbins_list), np.array(quantity_list)


def reformat_df(df, key, settings, keep=None, group_bayesian=False):
    """Change column names to format used by performance functions

    Args:
        df (pandas.DataFrame) (str): with columns [key_class*]
        keep (list) (str): list of keys to keep apart from ["target","SNTYPE","SNID","SIM_REDSHIFT_CMB"} if available

    Returns:
        df (pandas.DataFrame) (str): with columns [class*, predicted_target]
    """
    # Key will change depending if around max or complete lc
    if key != "all" and key != "probability":
        format_key = f"PEAKMJD{key}"
    else:
        format_key = key

    # new dataframe to save
    tmp_df = pd.DataFrame()
    # get class columns and reformat
    ori_class_columns = [k for k in df.keys() if f"{format_key}_class" in k]
    tmp_df[[k.split("_")[-1] for k in ori_class_columns]] = df[ori_class_columns].copy()
    # get the rest of features
    cols_to_keep = [
        k
        for k in ["target", settings.sntype_var, "SNID", "SIM_REDSHIFT_CMB"]
        if k in df.keys()
    ]
    if keep:
        cols_to_keep += keep
    tmp_df[cols_to_keep] = df[cols_to_keep].copy()
    # if variational or bayesian we can get the median prediction only
    if group_bayesian:
        tmp_df = tmp_df.groupby("SNID", as_index=False).median()
    # protect against NaNs (e.g. too early classifications)
    tmp_df = tmp_df[~np.isnan(tmp_df["class0"])]
    # set predicted target to max value in columns
    tmp_df["predicted_target"] = (
        tmp_df[[k for k in tmp_df.keys() if "class" in k]]
        .idxmax(axis=1)
        .str.strip("class")
        .astype(int)
    )
    return tmp_df


def create_latex_accuracy_singletable(df, outname, title):
    """Create latex tables from summary stats selection
    Table of metric by redshift given and early classification

    Args:
        df (pandas.DataFrame) : dataframe to convert
        outname (str) : name of latex table
        (optional) metric (str): which metric to output, default is accuracy

    Returns:
        Latex table: to write

    """

    # Get redshifts and accuracies in good format
    df = df.copy()
    df["redshift"] = df["model_name_noseed"].apply(
        lambda x: re.search(r"(?<=R\_)[A-Za-z]+", x).group()
    )
    metric = "accuracy"
    list_keys = ["-2", "0", "+2", "all"]
    for k in list_keys:
        df[k] = (
            "$"
            + df[f"{k}_{metric}_mean"].round(2).map(str)
            + " \\pm "
            + df[f"{k}_{metric}_std"].round(2).map(str)
            + "$"
        )

    to_write = df.to_latex(index=False, columns=["redshift"] + list_keys, escape=False)

    # Formatting table for MNRAS
    title_to_use = "\n\\multicolumn{5}{c}{" + title + "} \\\\\n"
    to_write = to_write.replace("\n", title_to_use, 1)
    to_write = (
        to_write.replace("toprule", "hline")
        .replace("midrule", "hline")
        .replace("bottomrule", "hline")
        .replace("{lllll}", "{l  cccc }")
    )

    return to_write



import torch.nn as nn
import torch
from supernnova.utils import logging_utils as lu
from supernnova.utils.swag_utils import SwagModel
from supernnova.training import bayesian_rnn
from supernnova.training import bayesian_rnn_2
from supernnova.training import variational_rnn
from supernnova.training import vanilla_rnn
import os
import h5py
import json
import pickle
import numpy as np
from tqdm import tqdm
from pathlib import Path

from sklearn import metrics
import matplotlib.pyplot as plt

plt.switch_backend("agg")


def normalize_arr(arr, settings):
    """Normalize array before input to RNN

    - Log transform
    - Mean and std dev normalization

    Args:
        arr (np.array) array to normalize
        settings (ExperimentSettings): controls experiment hyperparameters

    Returns:
        (np.array) the normalized array
    """

    if settings.norm == "none":
        return arr

    else:
        arr_min = settings.arr_norm[:, 0]
        arr_mean = settings.arr_norm[:, 1]
        arr_std = settings.arr_norm[:, 2]

        arr_to_norm = arr[:, settings.idx_features_to_normalize]
        # clipping
        arr_to_norm = np.clip(arr_to_norm, arr_min, np.inf)

        if "cosmo" not in settings.norm:
            # normalize using global norm
            arr_normed = np.log(arr_to_norm - arr_min + 1e-5)
            arr_normed = (arr_normed - arr_mean) / arr_std

        elif settings.norm == "cosmo":
            # normalize all lcs to 1 (fluxes), maintain color info
            # time is normalized as global norm
            arr_normed_cosmo = arr_to_norm
            # Flux + flux err normalization
            arr_normed_cosmo[:, :-1] = (
                arr_normed_cosmo[:, :-1] / arr_normed_cosmo[:, :-1].max()
            )
            # time normalization (log norm)
            tmp_cosmo = np.log(arr_to_norm[:, -1] - arr_min[-1] + 1e-5)
            arr_normed_cosmo[:, -1] = (tmp_cosmo - arr_mean[-1]) / arr_std[-1]
            # all together
            arr_normed = arr_normed_cosmo

        elif settings.norm == "cosmo_quantile":
            # normalize all lcs to 1 (using quantile not max for outliers)
            # maintain color info
            # time is normalized as global norm
            arr_normed_cosmo = arr_to_norm
            # Flux + flux err normalization
            quant_max = np.quantile(arr_normed_cosmo[:, :-1], 0.99)
            # clip quant_min and max
            arr_normed_cosmo[:, :-1] = arr_normed_cosmo[:, :-1] / quant_max

            # time normalization (log norm)
            tmp_cosmo = np.log(arr_to_norm[:, -1] - arr_min[-1] + 1e-5)
            arr_normed_cosmo[:, -1] = (tmp_cosmo - arr_mean[-1]) / arr_std[-1]
            # all together
            arr_normed = arr_normed_cosmo

        arr[:, settings.idx_features_to_normalize] = arr_normed

    return arr


def unnormalize_arr(arr, settings):
    """UnNormalize array

    Args:
        arr (np.array) array to normalize
        settings (ExperimentSettings): controls experiment hyperparameters

    Returns:
        (np.array) the normalized array
    """

    if settings.norm == "none":
        return arr

    arr_min = settings.arr_norm[:, 0]
    arr_mean = settings.arr_norm[:, 1]
    arr_std = settings.arr_norm[:, 2]

    if "cosmo" in settings.norm:
        # onyl unnormalize time
        arr_to_unnorm = arr[:, settings.idx_features_to_normalize[-1]]

        arr_to_unnorm = arr_to_unnorm * arr_std[-1] + arr_mean[-1]
        arr_unnormed = np.exp(arr_to_unnorm) + arr_min[-1] - 1e-5

        arr[:, settings.idx_features_to_normalize[-1]] = arr_unnormed

    else:
        arr_to_unnorm = arr[:, settings.idx_features_to_normalize]

        arr_to_unnorm = arr_to_unnorm * arr_std + arr_mean
        arr_unnormed = np.exp(arr_to_unnorm) + arr_min - 1e-5

        arr[:, settings.idx_features_to_normalize] = arr_unnormed

    return arr


def fill_data_list(
    idxs, arr_data, arr_target, arr_SNID, settings, n_features, desc, test=False
):
    """Utility to create a list of data tuples used as inputs to RNN model

    The ``settings`` object specifies which feature are selected

    Args:
        idxs (np.array or list): idx of data point to select
        arr_data (np.array): features
        arr_target (np.array): target
        arr_SNID (np.array): lightcurve unique ID
        settings (ExperimentSettings): controls experiment hyperparameters
        n_features (int): total number of features in arr_data
        desc (str): message to display while loading
        test (bool): If True: add more data to the list, as it is required at test time.
            Default: ``False``

    Returns:
        (list) the list of data tuples
    """

    list_data = []

    if desc == "":
        iterator = idxs
    else:
        iterator = tqdm(idxs, desc=desc, ncols=100)

    for i in iterator:
        X_all = arr_data[i].reshape(-1, n_features)
        target = int(arr_target[i])
        lc = str(arr_SNID[i])

        # Keep an unnormalized copy of the data (for test and display)
        X_ori = X_all.copy()[:, settings.idx_features]

        # check if normalization converges
        # using clipping in case of min<model_min
        X_clip = X_all.copy()
        X_clip = np.clip(
            X_clip[:, settings.idx_features_to_normalize],
            settings.arr_norm[:, 0],
            np.inf,
        )
        X_all[:, settings.idx_features_to_normalize] = X_clip

        if "cosmo" not in settings.norm:
            X_tmp = unnormalize_arr(normalize_arr(X_all.copy(), settings), settings)
            assert np.all(
                np.all(np.isclose(np.ravel(X_all), np.ravel(X_tmp), atol=1e-1))
            )

        # Normalize features that need to be normalized
        X_normed = X_all.copy()
        X_normed_tmp = normalize_arr(X_normed, settings)
        # Select features as specified by the settings
        X_normed = X_normed_tmp[:, settings.idx_features]

        if test is True:
            list_data.append((X_normed, target, lc, X_all, X_ori))
        else:
            list_data.append((X_normed, target, lc))

    return list_data


def load_HDF5(settings, test=False):
    """Load data from HDF5

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        test (bool): If True: load data for test. Default: ``False``

    Returns:
        list_data_test (list) test data tuples if test is True

        or

        Tuple containing
            - list_data_train (list): training data tuples
            - list_data_val (list): validation data tuples
    """
    file_name = f"{settings.processed_dir}/database.h5"
    lu.print_green(f"Loading {file_name}")

    with h5py.File(file_name, "r") as hf:

        list_data_train = []
        list_data_val = []

        config_name = f"{settings.source_data}_{settings.nb_classes}classes"

        dataset_split_key = f"dataset_{config_name}"
        target_key = f"target_{settings.nb_classes}classes"

        if test:
            # ridiculous failsafe in case we have different classes in dataset/model
            # we will always have 2 classes
            try:
                idxs_test = np.where(hf[dataset_split_key][:] == 2)[0]
            except Exception:
                idxs_test = np.where(hf["dataset_photometry_2classes"][:] != 100)[0]
        else:
            idxs_train = np.where(hf[dataset_split_key][:] == 0)[0]
            idxs_val = np.where(hf[dataset_split_key][:] == 1)[0]
            idxs_test = np.where(hf[dataset_split_key][:] == 2)[0]

            # Shuffle for good measure
            np.random.shuffle(idxs_train)
            np.random.shuffle(idxs_val)
            np.random.shuffle(idxs_test)

            idxs_train = idxs_train[: int(settings.data_fraction * len(idxs_train))]

        n_features = hf["data"].attrs["n_features"]
        training_features_data = hf["features"][:].astype(str)
        training_features = settings.training_features

        # DES5YR backward compatibility
        # not very robust
        if not set(training_features) <= set(training_features_data) and any(
            "DES-" in k for k in training_features_data
        ):
            training_features_data = [
                k.replace("DES-", "") for k in training_features_data
            ]

        # check if all model features are in db
        assert set(training_features) <= set(training_features_data)
        # check if all db features are in configuration
        if not set(settings.training_features_to_normalize) <= set(training_features):
            lu.print_red("Some features in the model are not in the configuration")
            lu.print_red("Database does not contain all training features")
            print("Database features:", training_features_data)
            print("Training features:", settings.training_features_to_normalize)
            lu.print_red(">>> Hint: check your provided --list_filters")
            raise ValueError

        lu.print_green("Features used", " ".join(training_features))

        try:
            settings.data_types_training = hf["data_types_training"][:].astype(str)
        except Exception:
            pass

        arr_data = hf["data"][:]
        if test:
            # ridiculous failsafe in case we have different classes in dataset/model
            # we will always have 2 classes
            try:
                arr_target = hf[target_key][:]
            except Exception:
                arr_target = hf["target_2classes"][:]
        else:
            arr_target = hf[target_key][:]
        arr_SNID = hf["SNID"][:].astype(str)

        if test is True:
            return fill_data_list(
                idxs_test,
                arr_data,
                arr_target,
                arr_SNID,
                settings,
                n_features,
                "Loading Test Set",
                test,
            )
        else:

            list_data_train = fill_data_list(
                idxs_train,
                arr_data,
                arr_target,
                arr_SNID,
                settings,
                n_features,
                "Loading Training Set",
            )
            list_data_val = fill_data_list(
                idxs_val,
                arr_data,
                arr_target,
                arr_SNID,
                settings,
                n_features,
                "Loading Validation Set",
            )

        return list_data_train, list_data_val


def get_model(settings, input_size):
    """Create RNN model

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        input_size (int): dimension of the input data

    Returns:
        (torch.nn Model) pytorch model
    """

    if settings.model == "vanilla":
        rnn = vanilla_rnn.VanillaRNN
    elif settings.model == "variational":
        rnn = variational_rnn.VariationalRNN
    elif settings.model == "bayesian":
        rnn = bayesian_rnn.BayesianRNN
    elif settings.model == "bayesian_2":
        rnn = bayesian_rnn_2.BayesianRNN
    elif settings.model == "swag":
        rnn = vanilla_rnn.VanillaRNN

    rnn = rnn(input_size, settings)
    
    if settings.model == "swag":
        # Wrap the model in SwagModel if swag is True
        rnn = SwagModel( rnn)

    if not settings.no_dump:
        print(rnn)

    return rnn


def get_optimizer(settings, model):
    """Create gradient descent optimizer

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        model (torch.nn Model): the pytorch model

    Returns:
        (torch.optim) the gradient descent optimizer
    """

    optimizer = torch.optim.Adam(
        model.parameters(),
        lr=settings.learning_rate,
        weight_decay=settings.weight_decay,
    )

    return optimizer


def get_data_batch(list_data, idxs, settings, max_lengths=None, OOD=None):
    """Create a batch in a deterministic way

    Args:
        list_data: (list) tuples of (X, target, lightcurve_ID)
        idxs: (array / list) indices of batch element in list_data
        settings (ExperimentSettings): controls experiment hyperparameters
        max_length (int): Maximum light curve length to be used Default: ``None``.
        OOD (str): Whether to modify data to create out of distribution data to be used Default: ``None``.

    Returns:
        Tuple containing
            - packed_tensor (torch PackedSequence): the packed features
            - X_tensor (torch Tensor): the features
            - target_tensor (torch Tensor): the target
    """

    list_len = []
    list_batch = []

    for pos, i in enumerate(idxs):
        X, target, *_ = list_data[i]
        # X is (L, D)
        if OOD is not None:
            # Make a copy to be sure we do not alter the original data
            X = X.copy()
        if OOD == "reverse":
            # For OOD test, reverse the sequence
            X = np.ascontiguousarray(X[::-1])
        elif OOD == "shuffle":
            # For OOD test, shuffle X
            p = np.random.permutation(X.shape[0])
            X = X[p]
        elif OOD == "sin":
            # For OOD test, set sine values to fluxes
            arr_flux = X[:, settings.idx_flux]
            arr_fluxerr = X[:, settings.idx_fluxerr]

            X_unnorm = unnormalize_arr(X.copy(), settings)
            arr_delta_time = X_unnorm[:, settings.idx_delta_time]
            arr_MJD = np.cumsum(arr_delta_time, axis=0)

            # Sine oscillations with 30 day period
            X[:, settings.idx_flux] = np.sin(arr_MJD * 2 * np.pi / 30) * np.max(
                arr_flux, axis=0, keepdims=True
            )
            X[:, settings.idx_fluxerr] = np.random.uniform(
                arr_fluxerr.min(), arr_fluxerr.max(), size=arr_fluxerr.shape
            )
        elif OOD == "random":
            # For OOD test, set random fluxes and errors
            arr_flux = X[:, settings.idx_flux]
            arr_fluxerr = X[:, settings.idx_fluxerr]

            X[:, settings.idx_flux] = np.random.uniform(
                arr_flux.min(), arr_flux.max(), size=arr_flux.shape
            )
            X[:, settings.idx_fluxerr] = np.random.uniform(
                arr_fluxerr.min(), arr_fluxerr.max(), size=arr_fluxerr.shape
            )

        if max_lengths is not None:
            assert settings.random_length is False
            assert settings.random_redshift is False
            X = X[: max_lengths[pos]]
        if settings.random_length:
            random_length = np.random.randint(3, X.shape[0] + 1)
            X = X[:random_length]
        if settings.redshift == "zspe" and settings.random_redshift:
            if np.random.binomial(1, 0.5) == 0:
                X[:, settings.idx_specz] = -1
        input_dim = X.shape[1]
        list_len.append(X.shape[0])
        list_batch.append((X, target))

    # Get indices to sort the batch by sequence size (needed to use packed sequences in pytorch)
    # Sequences should be arranged in decreasing length
    idx_sort = np.argsort(list_len)[::-1]
    idxs_rev_sort = np.argsort(idx_sort)  # these indices revert the sort
    max_len = list_len[idx_sort[0]]
    X_tensor = torch.zeros((max_len, len(idxs), input_dim))
    list_target = []
    lengths = []
    # Assign values for the tensor
    for i, idx in enumerate(idx_sort):
        X, target = list_batch[idx]
        try:
            X_tensor[: X.shape[0], i, :] = torch.FloatTensor(X)
        except Exception:
            X_tensor[: X.shape[0], i, :] = torch.FloatTensor(
                torch.from_numpy(np.flip(X, axis=0).copy().astype(np.float32))
            )
        list_target.append(target)
        lengths.append(list_len[idx])

    # Move data to GPU if required
    if settings.use_cuda:
        X_tensor = X_tensor.cuda()
        target_tensor = torch.LongTensor(list_target).cuda()

    else:
        X_tensor = X_tensor
        target_tensor = torch.LongTensor(list_target)

    # Create a packed sequence
    packed_tensor = nn.utils.rnn.pack_padded_sequence(X_tensor, lengths)

    return packed_tensor, X_tensor, target_tensor, idxs_rev_sort


def train_step(
    settings,
    rnn,
    packed_tensor,
    target_tensor,
    criterion,
    optimizer,
    batch_size,
    num_batches,
):
    """Full training step : Forward and Backward pass

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        rnn (torch.nn Model): pytorch model to train
        packed_tensor (torch PackedSequence): input tensor in packed form
        target_tensor (torch Tensor): target tensor
        criterion (torch loss function): loss function to optimize
        optimizer (torch optim): the gradient descent optimizer
        batch_size (int): batch size
        num_batches (int): number of minibatches to scale KL cost in Bayesian
    """

    # Set NN to train mode (deals with dropout and batchnorm)
    rnn.train()

    # Zero out the gradients
    optimizer.zero_grad()

    # Forward pass
    output = rnn(packed_tensor)
    loss = criterion(output.squeeze(), target_tensor)
    # Special case for BayesianRNN, need to use KL loss
    if isinstance(rnn, bayesian_rnn.BayesianRNN):
        loss = loss + rnn.kl / (num_batches * batch_size)
    else:
        loss = criterion(output.squeeze(), target_tensor)

    # Backward pass
    loss.backward()
    optimizer.step()
    return loss


def eval_step(rnn, packed_tensor):
    """Eval step: Forward pass only

    Args:
        rnn (torch.nn Model): pytorch model to evaluate
        packed_tensor (torch PackedSequence): input tensor in packed form

    Returns:
        output (torch Tensor): output of rnn
    """

    # Set NN to eval mode (deals with dropout and batchnorm)
    rnn.eval()

    # Forward pass
    output = rnn(packed_tensor)
    return output


def eval_step_swag(model: SwagModel, scale: float, cov: bool, packed_tensor):
    """Draw a sample of SwagModel and evaluate: Forward pass only

    Args:
        model (SwagModel): SwagModel
        scale (float): The scale parameter for covariance
        cov (bool): If True, enable calculating low-rank covariance
        packed_tensor (torch PackedSequence): Input tensor in packed form

    Returns:
        output (torch Tensor): Output of a sample model
    """
    # draw a sample of SwagModel
    sample_model = model.sample(scale, cov)

    # Set NN to eval mode
    sample_model.eval()

    # Forward pass
    output = sample_model(packed_tensor)

    # for debug only
    if torch.isnan(sample_model(packed_tensor)).any():
        print("sample_model output contains NAN value")
        breakpoint()

    if torch.isnan(output).any():
        print("sum of sample model output contain NAN value")
        breakpoint()

    return output


def plot_loss(d_train, d_val, epoch, settings):
    """Plot loss curves

    Plot training and validation logloss

    Args:
        d_train (dict of arrays): training log losses
        d_val (dict of arrays): validation log losses
        epoch (int): current epoch
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    for key in d_train.keys():

        plt.figure()
        plt.plot(d_train["epoch"], d_train[key], label="Train %s" % key.title())
        plt.plot(d_val["epoch"], d_val[key], label="Val %s" % key.title())
        plt.legend(loc="best", fontsize=18)
        plt.xlabel("Step", fontsize=22)
        plt.tight_layout()
        plt.savefig(
            Path(settings.models_dir)
            / f"{settings.pytorch_model_name}"
            / f"train_and_val_{key}_{settings.pytorch_model_name}.png"
        )
        plt.close()
        plt.clf()


def get_evaluation_metrics(
    settings, list_data, model, sample_size=None, swag_sampling=False
):
    """Compute evaluation metrics on a list of data points

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        list_data (list): contains data to evaluate
        model (torch.nn Model): pytorch model
        sample_size (int): subset of the data to use for validation. Default: ``None``
        swag_sampling (bool): if True, enable to draw swag samplings. Default: ``False``

    Returns:
        d_losses (dict) maps metrics to their computed value
    """

    # Validate
    list_pred = []
    list_target = []
    list_kl = []
    num_elem = len(list_data)
    num_batches = num_elem // min(num_elem // 2, settings.batch_size)
    list_batches = np.array_split(np.arange(num_elem), num_batches)

    # If required, pick a subset of list batches at random
    if sample_size:
        batch_idxs = np.random.permutation(len(list_batches))
        num_batches = sample_size // min(sample_size // 2, settings.batch_size)
        batch_idxs = batch_idxs[:num_batches]
        list_batches = [list_batches[batch_idx] for batch_idx in batch_idxs]

    for batch_idxs in list_batches:
        random_length = settings.random_length
        settings.random_length = False
        packed_tensor, X_tensor, target_tensor, idxs_rev_sort = get_data_batch(
            list_data, batch_idxs, settings
        )
        settings.random_length = random_length

        if swag_sampling:
            scale = settings.swag_scale
            cov = True
            if settings.swag_no_lr_cov:
                cov = False
            output = eval_step_swag(model, scale, cov, packed_tensor)
        else:
            output = eval_step(model, packed_tensor)

        # for debug only
        if torch.isnan(output).any():
            print("output contains NAN value")
            breakpoint()

        if "bayesian" in settings.pytorch_model_name:
            list_kl.append(model.kl.detach().cpu().item())

        # Apply softmax
        pred_proba = nn.functional.softmax(output, dim=1)
        # for debug only
        if torch.isnan(pred_proba).any():
            print("pred_proba contains NAN vlaue")
            breakpoint()
        # Convert to numpy array
        pred_proba = pred_proba.data.cpu().numpy()
        target_numpy = target_tensor.data.cpu().numpy()

        # Revert sort
        pred_proba = pred_proba[idxs_rev_sort]
        target_numpy = target_numpy[idxs_rev_sort]

        list_pred.append(pred_proba)
        list_target.append(target_numpy)
    targets = np.concatenate(list_target, axis=0)
    preds = np.concatenate(list_pred, axis=0)
    # for debug only
    if np.isnan(preds).any():
        print("preds contains NAN value")
        breakpoint()
    # Check outputs size
    assert len(targets.shape) == 1
    assert len(preds.shape) == 2

    if settings.nb_classes == 2 and len(set(targets)) > 1:
        auc = metrics.roc_auc_score(targets, preds[:, 1])
    else:
        # Can't compute AUC for more than 2 classes
        auc = None
    acc = metrics.accuracy_score(targets, np.argmax(preds, 1))
    targets_2D = np.zeros((targets.shape[0], settings.nb_classes))
    for i in range(targets.shape[0]):
        targets_2D[i, targets[i]] = 1
    log_loss = metrics.log_loss(targets_2D, preds)

    d_losses = {"AUC": auc, "Acc": acc, "loss": log_loss}

    if len(list_kl) != 0:
        d_losses["KL"] = np.mean(list_kl)

    return d_losses


def get_loss_string(d_losses_train, d_losses_val):
    """Obtain a loss string to display training progress

    Args:
        d_losses_train (dict): maps {metric:value} for the training data
        d_losses_val (dict): maps {metric:value} for the validation data

    Returns:
        loss_str (str): the loss string to display
    """

    loss_str = "/".join(d_losses_train.keys())

    loss_str += " [T]: " + "/".join(
        [
            f"{value:.3g}" if (value is not None and key != "epoch") else "NA"
            for (key, value) in d_losses_train.items()
        ]
    )
    loss_str += " [V]: " + "/".join(
        [
            f"{value:.3g}" if (value is not None and key != "epoch") else "NA"
            for (key, value) in d_losses_val.items()
        ]
    )

    return loss_str


def save_training_results(settings, d_monitor, training_time):
    """Obtain a loss string to display training progress

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        d_monitor (dict): maps {metric:value}
        training_time (float): amount of time training took

    Returns:
        loss_str (str): the loss string to display
    """

    d_results = {"training_time": training_time}
    for key in ["AUC", "Acc"]:
        if key == "AUC" and settings.nb_classes > 2:
            d_results[key] = -1
        else:
            d_results[key] = max(d_monitor[key]) if None not in d_monitor[key] else None
    d_results["loss"] = min(d_monitor["loss"])

    d_results["data_types_training"] = np.array2string(settings.data_types_training)

    try:
        with open(Path(settings.rnn_dir) / "training_log.json", "r") as f:
            d_out = json.load(f)
    except Exception:
        d_out = {}

    with open(Path(settings.rnn_dir) / "training_log.json", "w") as f:
        d_out.update({settings.pytorch_model_name: d_results})
        json.dump(d_out, f)


#######################
# RandomForest Utils
#######################


def save_randomforest_model(settings, clf):
    """Save RandomForest model

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        clf (RandomForestClassifier): RandomForest model
    """

    filename = f"{settings.rf_dir}/{settings.randomforest_model_name}.pickle"
    with open(filename, "wb") as f:
        pickle.dump(clf, f)
    lu.print_green("Saved model")


def load_randomforest_model(settings, model_file=None):
    """Load RandomForest model

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        model_file (str): path to saved randomforest model. Default: ``None``

    Returns:
        (RandomForestClassifier) RandomForest model
    """

    if model_file is None:
        model_file = f"{settings.rf_dir}/{settings.randomforest_model_name}.pickle"
    assert os.path.isfile(model_file)
    with open(model_file, "rb") as f:
        clf = pickle.load(f)
    lu.print_green("Loaded model")

    return clf


def train_and_evaluate_randomforest_model(clf, X_train, y_train, X_val, y_val):
    """Train a RandomForestClassifier and evaluate AUC, precision, accuracy
    on a validation set

    Args:
        clf (RandomForestClassifier): RandomForest model to fit and evaluate
        X_train (np.array): the training features
        y_train (np.array): the training target
        X_val (np.array): the validation features
        y_val (np.array): the validation target
    """
    lu.print_green("Fitting RandomForest...")
    clf = clf.fit(X_train, y_train)
    lu.print_green("Fitting complete")

    # Evaluate our classifier
    probas_ = clf.predict_proba(X_val)
    # Compute AUC and precision
    fpr, tpr, thresholds = metrics.roc_curve(y_val, probas_[:, 1])
    roc_auc = metrics.auc(fpr, tpr)
    pscore = metrics.precision_score(y_val, clf.predict(X_val), average="binary")
    lu.print_green("Validation AUC", roc_auc)
    lu.print_green("Validation precision score", pscore)

    lu.print_green(
        "Train data accuracy",
        100 * (sum(clf.predict(X_train) == y_train)) / X_train.shape[0],
    )
    lu.print_green(
        "Val data accuracy", 100 * (sum(clf.predict(X_val) == y_val)) / X_val.shape[0]
    )

    return clf


class StopOnPlateau(object):
    """
    Detect plateau on accuracy (or any metric)
    If chosen, will reduce learning rate of optimizer once in the Plateau

    .. code: python

          plateau_accuracy = tu.StopOnPlateau()
          for epoch in range(10):
              ... get metric ...
              plateau = plateau_accuracy.step(metric_value)
              if plateau is True:
                   break

    Args:
        patience (int): number of epochs to wait, after which we decrease the LR
            if the validation loss is plateauing
        reduce_lr-on_plateau (bool): If True, reduce LR after loss has not improved
            in the last patience epochs
        max_learning_rate_reduction (float): max factor by which to reduce the learning rate
    """

    def __init__(
        self, patience=10, reduce_lr_on_plateau=False, max_learning_rate_reduction=3
    ):

        self.patience = patience
        self.best = 0.0
        self.num_bad_epochs = 0
        self.is_better = None
        self.last_epoch = -1
        self.list_metric = []
        self.reduce_lr_on_plateau = reduce_lr_on_plateau
        self.max_learning_rate_reduction = max_learning_rate_reduction
        self.learning_rate_reduction = 0

    def step(self, metric_value, optimizer=None, epoch=None):
        current = metric_value
        if epoch is None:
            epoch = self.last_epoch = self.last_epoch + 1
        self.last_epoch = epoch

        # Are we under .05 std in accuracy on the last 10 epochs
        self.list_metric.append(current)
        if len(self.list_metric) > 10:
            self.list_metric = self.list_metric[-10:]
            # are we in a plateau?
            # accuracy is not in percentage, so two decimal numbers is actually 4 in this notation
            if np.array(self.list_metric).std() < 0.0005:
                print("Has reached a learning plateau with", current, "\n")
                if optimizer is not None and self.reduce_lr_on_plateau is True:
                    print(
                        "Reducing learning rate by factor of ten",
                        self.learning_rate_reduction,
                        "\n",
                    )
                    for param in optimizer.param_groups:
                        param["lr"] = param["lr"] / 10.0
                    self.learning_rate_reduction += 1
                    if self.learning_rate_reduction == self.max_learning_rate_reduction:
                        return True
                else:
                    return True
            else:
                return False



import matplotlib.pylab as plt

# Plotting styles
ALL_COLORS = [
    "maroon",
    "darkorange",
    "royalblue",
    "indigo",
    "black",
    "red",
    "purple",
    "maroon",
    "darkorange",
    "royalblue",
    "indigo",
]
BI_COLORS = ["darkorange", "royalblue"]
CONTRAST_COLORS = ["darkorange", "indigo"]
MARKER_DIC = {"vanilla": "s"}
FILL_DIC = {"None": "none", "zpho": "bottom", "zspe": "full"}
MARKER_LIST = ["o", "o", "v", "v", "^", "^", ">", ">", "<", "<", "s", "s", "D", "D"]
CMAP = plt.cm.YlOrBr
LINE_STYLE = ["-", "-", "-", "-", ":", ":", ":", ":", "-.", "-.", "-.", "-."]
FILTER_COLORS = {
    "z": "maroon",
    "i": "darkorange",
    "r": "royalblue",
    "g": "indigo",
    "u": "purple",
    "Y": "red",
}
PATTERNS = ["", "", "", "", "", ".", ".", ".", "."]


def get_model_visualization_name(model_name):

    if "bayesian" in model_name or "BBB" in model_name:
        return "BBB RNN"
    if "variational" in model_name:
        return "Variational RNN"
    if "vanilla" in model_name or "baseline" in model_name:
        return "Baseline RNN"



import os
from pathlib import Path
import h5py
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

plt.switch_backend("agg")


def plot_lightcurves(df, SNIDs, settings):
    """Utility for gridspec of lightcruves

    Args:
        df (pandas.DataFrame): dataframe holding the data
        SNIDs (np.array or list): holds lightcurve ids
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    plt.figure(figsize=(20, 10))
    plt.suptitle("Sample of SN Ia light curves")
    gs = gridspec.GridSpec(4, 4, hspace=0.4)

    for i in range(16):
        ax = plt.subplot(gs[i])

        SNID = SNIDs[i]
        df_temp = df.loc[SNID]

        # Prepare plotting data in a dict
        d = {
            flt: {"FLUXCAL": [], "FLUXCALERR": [], "MJD": []}
            for flt in settings.list_filters
        }

        current_time = 0
        for idx in range(len(df_temp)):
            flt = df_temp.FLT.values[idx]
            d[flt]["FLUXCAL"].append(df_temp.FLUXCAL.values[idx])
            d[flt]["FLUXCALERR"].append(df_temp.FLUXCALERR.values[idx])
            current_time += df_temp.delta_time.values[idx]
            d[flt]["MJD"].append(current_time)

        for flt in d.keys():
            time = d[flt]["MJD"]
            # Only plot a time series if it's non empty
            if len(time) > 0:
                flux = d[flt]["FLUXCAL"]
                fluxerr = d[flt]["FLUXCALERR"]
                ax.errorbar(time, flux, yerr=fluxerr, label=f"Filter {flt}")

        ax.set_title(SNID, fontsize=18)
        ax.set_aspect("auto")
    ax.legend(loc="best")

    plt.savefig(Path(settings.explore_dir) / "sample_lightcurves.png")


def plot_random_preprocessed_lightcurves(settings, SNIDs):
    """Plot lightcurves specified by SNID_idxs from the
    preprocessed, pickled database

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        SNIDs (list): list of SN lightcurve IDs to plot
    """

    list_files = [
        f for f in glob.glob(os.path.join(settings.preprocessed_dir, "*_PHOT.pickle"))
    ]
    if len(list_files) == 0:
        raise FileNotFoundError(
            "No files found in the preprocessed directory. Use option --debug together with --data when generating database."
        )

    df = pd.concat(list(map(pd.read_pickle, list_files))).set_index("SNID")

    # Plot and save
    plot_lightcurves(df, SNIDs, settings)


def plot_lightcurves_from_hdf5(settings, SNID_idxs):
    """Plot lightcurves specified by SNID_idxs from the
    HDF5 database

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        SNID_idxs (list): list of SN lightcurve index to plot
    """

    with h5py.File(settings.hdf5_file_name, "r") as hf:

        features = hf["features"][:].astype(str)
        n_features = len(features)

        plt.figure(figsize=(20, 10))
        gs = gridspec.GridSpec(4, 4, hspace=0.4)

        for idx, SNID_idx in enumerate(SNID_idxs):

            ax = plt.subplot(gs[idx])

            SNID = hf["SNID"][SNID_idx]
            str(hf["PEAKMJD"][SNID_idx])
            PEAKMJDNORM = hf["PEAKMJDNORM"][SNID_idx]
            typ = hf[settings.sntype_var][SNID_idx]
            typ = settings.sntypes[str(typ)]
            data = hf["data"][SNID_idx].reshape(-1, n_features)

            df = pd.DataFrame(data, columns=features)

            non_filter_columns = [
                "FLUXCAL_g",
                "FLUXCAL_i",
                "FLUXCAL_r",
                "FLUXCAL_z",
                "FLUXCALERR_g",
                "FLUXCALERR_i",
                "FLUXCALERR_r",
                "FLUXCALERR_z",
                "delta_time",
                "HOSTGAL_PHOTOZ",
                "HOSTGAL_PHOTOZ_ERR",
                "HOSTGAL_SPECZ",
                "HOSTGAL_SPECZ_ERR",
            ]

            filter_columns = [
                c for c in df.columns.values if c not in non_filter_columns
            ]

            present_filters = df[filter_columns].transpose().idxmax().values
            list_present_filters = [set(f) for f in present_filters]

            max_y = -float("Inf")
            min_y = float("Inf")

            for FLT in settings.list_filters:
                idxs = np.array(
                    [i for i in range(len(df)) if FLT in list_present_filters[i]]
                )
                if len(idxs) == 0:
                    continue
                arr_flux = df[f"FLUXCAL_{FLT}"].values[idxs]
                arr_fluxerr = df[f"FLUXCALERR_{FLT}"].values[idxs]
                arr_time = df["delta_time"].cumsum().values[idxs]
                ax.errorbar(arr_time, arr_flux, yerr=arr_fluxerr, label=f"Filter {FLT}")

                if np.max(arr_flux) > max_y:
                    max_y = np.max(arr_flux)

                if np.min(arr_flux) < min_y:
                    min_y = np.min(arr_flux)

            ax.plot(
                [PEAKMJDNORM, PEAKMJDNORM], [min_y, max_y], color="k", linestyle="--"
            )
            ax.set_title(f"{SNID.decode('utf8')} -- {typ}", fontsize=18)
            ax.set_aspect("auto")
        ax.legend(loc="best")
        plt.savefig(Path(settings.explore_dir) / "sample_lightcurves_from_hdf5.png")


def visualize(settings):
    """Plot a random subset of lightcurves

    2 plots: one with preprocessed data and one with processed data
    The two plots should show the same data

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    # Check the data has been created
    settings.check_data_exists()

    # Set a random seed
    np.random.seed()

    with h5py.File(settings.hdf5_file_name, "r") as hf:
        SNID_idxs = np.random.permutation(hf["SNID"].shape[0])[:16]
        SNIDs = hf["SNID"][:][SNID_idxs]
    SNIDs = [i for i in np.array([k for k in SNIDs]).astype(str)]

    plot_random_preprocessed_lightcurves(settings, SNIDs)

    plot_lightcurves_from_hdf5(settings, SNID_idxs)



import sys
import re
import os.path as osp
import numpy as np
from pathlib import Path
from supernnova import conf
from supernnova.utils import logging_utils as lu

# Notice: This script serves as the entry point of the program. By design, only one action can be executed at a time.
# To optimize performance, heavy imports are deferred until their corresponding actions are invoked.

ALL_ACTIONS = ("make_data", "train_rnn", "validate_rnn", "show", "performance")

help_msg = """
Available commands:

    make_data        create dataset for ML training
    train_rnn        train RNN model
    validate_rnn     validate RNN model
    show             visualize different types of plot (requirement make_data)
    performance      get method performance and paper plots

Type snn <command> --help for usage help on a specific command.
For example, snn make_data --help will list all data creation options.
"""


def print_usage():
    print("Usage: {} <command> <options> <arguments>".format(osp.basename(sys.argv[0])))
    print(help_msg)


def get_action():
    """Pop first argument, check it is a valid action."""
    if len(sys.argv) <= 1:
        print_usage()
        sys.exit(1)
    if sys.argv[1] not in ALL_ACTIONS:
        print_usage()
        sys.exit(1)

    return sys.argv.pop(1)


def get_plot_lcs(settings):
    from supernnova.visualization import early_prediction

    if settings.model_files is None:
        early_prediction.make_early_prediction(settings, nb_lcs=2, do_gifs=False)

    elif settings.model_files:
        for model_file in settings.model_files:
            model_settings = conf.get_settings_from_dump(settings, model_file)
            early_prediction.make_early_prediction(
                model_settings, nb_lcs=2, do_gifs=False
            )


def make_data_action(settings):
    from supernnova.data import make_dataset
    from supernnova.visualization import visualize

    # Validate command-line arguments
    # explore_lightcurves should be used with debug
    if settings.explore_lightcurves and not settings.debug:
        message = "--explore_lightcurves must be used with --debug"
        raise ValueError(message)

    # Build an HDF5 database
    make_dataset.make_dataset(settings)
    lu.print_blue("Finished constructing dataset")

    # Visualize lightcurves
    if settings.explore_lightcurves:
        visualize.visualize(settings)


def train_rnn_action(settings):
    from supernnova.training import train_rnn
    from supernnova.validation import validate_rnn, metrics
    from supernnova.visualization import early_prediction
    from supernnova.paper import superNNova_plots as sp

    # Validate command-line arguments
    if settings.swag:
        if settings.cyclic:
            lu.print_yellow("SWAG is not available with training cyclic")
        if settings.swag_start_epoch + 2 >= settings.nb_epoch:
            message = "(swag_start_epoch +2) must be smaller than nb_epoch"
            raise ValueError(message)

    # Train
    if settings.cyclic:
        train_rnn.train_cyclic(settings)
    else:
        train_rnn.train(settings)

    # Obtain predictions
    prediction_file_list = validate_rnn.get_predictions(settings)

    # Compute metrics
    for prediction_file in prediction_file_list:
        metrics.get_metrics_singlemodel(
            settings, prediction_file=prediction_file, model_type="rnn"
        )

    # Plot some lightcurves
    early_prediction.make_early_prediction(settings)

    lu.print_blue("Finished rnn training, validating, testing and plotting lcs")

    if settings.calibration:
        sp.plot_calibration(settings, prediction_files=prediction_file_list)


def validate_rnn_action(settings):
    from supernnova.validation import validate_rnn, metrics
    from supernnova.visualization import prediction_distribution
    from supernnova.paper import superNNova_plots as sp

    if settings.model_files is None:
        prediction_file_list = validate_rnn.get_predictions(settings)
        # Compute metrics
        for prediction_file in prediction_file_list:
            metrics.get_metrics_singlemodel(
                settings, prediction_file=prediction_file, model_type="rnn"
            )
        pf = prediction_file_list
    else:
        prediction_files = []
        for model_file in settings.model_files:
            # Restore model settings
            model_settings = conf.get_settings_from_dump(settings, model_file)
            if settings.num_inference_samples != model_settings.num_inference_samples:
                model_settings.num_inference_samples = settings.num_inference_samples
            # Get predictions
            prediction_file_list = validate_rnn.get_predictions(
                model_settings, model_file=model_file
            )
            prediction_files.extend(prediction_file_list)
            # Compute metrics
            for prediction_file in prediction_file_list:
                metrics.get_metrics_singlemodel(
                    model_settings,
                    prediction_file=prediction_file,
                    model_type="rnn",
                )
        pf = prediction_files

    if settings.plot_lcs:
        get_plot_lcs(settings)

    if settings.calibration:
        sp.plot_calibration(settings, prediction_files=pf)

    if settings.plot_prediction_distribution:
        prediction_distribution.plot_prediction_distribution(settings)

    if settings.speed:
        validate_rnn.get_predictions_for_speed_benchmark(settings)


def show_action(settings):
    from supernnova.visualization import prediction_distribution
    from supernnova.paper import superNNova_plots as sp

    # Validate command-line arguments
    # prediction_files should be provided for calibration
    if settings.calibration and not settings.prediction_files:
        message = "--calibration must be used with --prediction_files"
        raise ValueError(message)

    if settings.plot_lcs:
        get_plot_lcs(settings)

    if settings.plot_prediction_distribution:
        prediction_distribution.plot_prediction_distribution(settings)

    if settings.calibration:
        sp.plot_calibration(settings)


def performance_action(settings):
    from supernnova.validation import validate_rnn, metrics

    # Validate command-line arguments
    # Need to provide prediction files when calculating metrics
    if settings.metrics and not settings.prediction_files:
        message = "--metrics should be used with --prediction_files"
        raise ValueError(message)

    if settings.metrics:
        for prediction_file in settings.prediction_files:
            # TODO: need to make sure only rnn model file is allowed in this step
            model_type = "rf" if "randomforest" in prediction_file else "rnn"
            metrics.get_metrics_singlemodel(
                conf.get_settings_from_dump(settings, prediction_file),
                prediction_file=prediction_file,
                model_type=model_type,
            )
        lu.print_blue("Finished computing metrics")

    # if settings.performance:
    #     from supernnova.utils import logging_utils
    #     metrics.aggregate_metrics(settings)
    #     lu.print_blue("Finished aggregating performance")
    #     # Stats and plots in paper
    #     st.SuperNNova_stats_and_plots(settings)
    #     lu.print_blue("Finished assembling paper performance")

    # Speed benchmarks
    if settings.speed:
        validate_rnn.get_predictions_for_speed_benchmark(settings)

    if settings.done_file:
        with open(Path(settings.done_file), "w") as the_file:
            the_file.write("SUCCESS\n")


def main():
    if len(sys.argv) == 2 and sys.argv[1] == "--help":
        print_usage()

    else:
        # Workaround for optparse limitation: insert -- before first negative
        # number found.
        negint = re.compile("-[0-9]+")
        for n, arg in enumerate(sys.argv):
            if negint.match(arg):
                sys.argv.insert(n, "--")
                break

        actions = {
            "make_data": make_data_action,
            "train_rnn": train_rnn_action,
            "validate_rnn": validate_rnn_action,
            "show": show_action,
            "performance": performance_action,
        }

        action = get_action()

        # Get config parameters
        settings = conf.get_settings(action)
        # setting random seeds
        np.random.seed(settings.seed)

        import torch

        torch.manual_seed(settings.seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(settings.seed)

        actions[action](settings)


if __name__ == "__main__":
    main()



import json
import torch
import numpy as np
import pandas as pd
from pathlib import Path
from supernnova.utils import data_utils
import supernnova.utils.logging_utils as lu
from supernnova.conf import get_norm_from_model
from supernnova.utils import experiment_settings
from supernnova.utils import training_utils as tu
from supernnova.validation.validate_rnn import get_batch_predictions
from supernnova.data.make_dataset import pivot_dataframe_single_from_df


def get_settings(model_file):
    """Define settings from model

    Args:
        model_file (str): complete name with path for model to be used
    Returns:
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    # Load model
    model_dir = Path(model_file).parent
    cli_file = model_dir / "cli_args.json"
    with open(cli_file, "r") as f:
        cli_args = json.load(f)
        # Unset general arguments
        for arg in [
            "data",
            "train_rnn",
            "validate_rnn",
            "explore_lightcurves",
            "dryrun",
            "metrics",
            # "performance",
            "calibration",
            "plot_lcs",
            "prediction_files",
        ]:
            cli_args[arg] = False

    # for on the fly predictions
    cli_args["no_dump"] = True
    cli_args["models_dir"] = Path(model_file).parent.as_posix()

    settings = experiment_settings.ExperimentSettings(cli_args)

    # Backward compatibility (hardcoded...)
    dic_missing_keys = {"sntype_var": "SNTYPE", "additional_train_var": None}
    for k, v in dic_missing_keys.items():
        if k not in cli_args.keys():
            cli_args[k] = v

    settings = experiment_settings.ExperimentSettings(cli_args)

    # load normalization from json dump
    settings = get_norm_from_model(model_file, settings)

    return settings


def format_data(df, settings):
    """Format data into SuperNNova-friendly format

    Args:
        df (pandas.DataFrame): dataframe with light-curves
        settings (ExperimentSettings): custom class to hold hyperparameters
    Returns:
        df (pandas.DataFrame): reformatted data

    """

    # fill missing columns with zeros
    # only if columns are not used for classification
    host_features_model = [k for k in settings.all_features if "HOST" in k]
    host_features_notindata = [k for k in host_features_model if k not in df.keys()]
    if (
        (len(host_features_notindata) > 0 and settings.redshift == "none")
        or (
            settings.redshift == "zspe"
            and "HOSTGAL_SPECZ" not in host_features_notindata
        )
        or settings.redshift == "zpho"
        and "HOSTGAL_PHOTOZ" not in host_features_notindata
    ):
        for hf in host_features_notindata:
            df[hf] = np.zeros(len(df))
    elif len(host_features_notindata) > 0:
        lu.print_red(
            f"Missing features needed for classification {host_features_notindata}"
        )
        raise ValueError

    # compute delta time
    df = data_utils.compute_delta_time(df)
    # fill dummies
    if "PEAKMJD" not in df.keys():
        df["PEAKMJD"] = np.zeros(len(df))

    # pivot
    df = pivot_dataframe_single_from_df(df, settings)

    # onehot
    # Fit a one hot encoder for FLT
    # to have the same onehot for all datasets
    tmp = pd.concat([pd.Series(settings.list_filters_combination), df["FLT"]])
    tmp_onehot = pd.get_dummies(tmp)
    # this is ok since it goes by length not by index (which I never reset)
    # beware: this requires index int!
    FLT_onehot = tmp_onehot[len(settings.list_filters_combination) :]

    df = pd.concat([df, FLT_onehot], axis=1)[settings.all_features]

    ordered_features = [k for k in settings.all_features if k in df.keys()]
    df = df[ordered_features]

    return df


def classify_lcs(df, model_file, device):
    """Obtain predictions for light-curves
    Args:
        df (DataFrame): light-curves to classify
        model_file (str): Path+name of model to use for predictions
        device (str): wehter to use cuda or cpu

    Returns:
        idx (list): light-curve indices after classification (they are resorted)
        preds (np.array): predictions for this model (shape= len(idx),model_nb_class)
    """

    # init
    settings = get_settings(model_file)
    settings.use_cuda = True if "cuda" in str(device) else False
    settings.idx_features_to_normalize = [
        i
        for (i, f) in enumerate(settings.training_features)
        if f in settings.training_features_to_normalize
    ]

    settings.random_length = False
    settings.random_redshift = False

    # Set the random seed manually for reproducibility.
    np.random.seed(settings.seed)
    torch.manual_seed(settings.seed)

    if "vanilla" in settings.pytorch_model_name:
        settings.num_inference_samples = 1

    # format data
    df = format_data(df, settings)

    # get packed data batches
    list_lcs = []

    for _, sel in df.groupby(level=0):
        X_all = sel.values
        # check if normalization converges
        # using clipping in case of min<model_min
        X_clip = X_all.copy()
        X_clip = np.clip(
            X_clip[:, settings.idx_features_to_normalize],
            settings.arr_norm[:, 0],
            np.inf,
        )
        X_all[:, settings.idx_features_to_normalize] = X_clip

        # Normalize features that need to be normalized
        X_normed = X_all.copy()
        X_normed = tu.normalize_arr(X_normed, settings)
        # format: data, target (filled with zeros), _
        settings.idx_features = [
            i
            for (i, _) in enumerate(settings.all_features)
            if _ in settings.training_features
        ]
        X_normed = X_normed[:, settings.idx_features]
        X_tmp = X_normed, 0, "dummy"

        list_lcs.append(X_tmp)

    packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
        list_lcs, np.arange(len(list_lcs)), settings
    )

    # load model
    rnn = tu.get_model(settings, len(settings.training_features))
    rnn_state = torch.load(
        model_file,
        map_location=lambda storage, loc: storage,
        weights_only=False,
    )
    try:
        rnn.load_state_dict(rnn_state)
    except Exception:
        if len(settings.training_features) < len(settings.all_features):
            lu.print_red("Model has less features than data (check model filters)")
        else:
            lu.print_red("Check correct model is chosen")
        raise ValueError
    rnn.to(device)
    rnn.eval()
    # obtain predictions
    list_preds = []
    for iter_ in range(settings.num_inference_samples):

        arr_preds, _ = get_batch_predictions(rnn, packed, target_tensor)

        # Rever sorting that occurs in get_batch_predictions
        arr_preds = arr_preds[idxs_rev_sort]

        list_preds.append(arr_preds)

    # B, inf_samples, nb_classes
    preds = np.stack(list_preds, axis=1)

    return df.index.unique(), preds
